from __future__ import annotations

import dataclasses
import inspect
import uuid
from collections import defaultdict
from itertools import count
from typing import TYPE_CHECKING, Any, Generic, Optional

from asgiref.sync import sync_to_async
from django.db.models import Q, Value
from django.db.models.manager import BaseManager
from graphql import GraphQLID, GraphQLObjectType

from undine import QueryType
from undine.dataclasses import OptimizationWithPagination, QuerySetMapWithPagination
from undine.exceptions import (
    GraphQLFieldNotNullableError,
    GraphQLModelNotFoundError,
    GraphQLNodeIDFieldTypeError,
    GraphQLNodeInterfaceMissingError,
    GraphQLNodeInvalidGlobalIDError,
    GraphQLNodeMissingIDFieldError,
    GraphQLNodeObjectTypeMissingError,
    GraphQLNodeQueryTypeMissingError,
    GraphQLNodeTypeNotObjectTypeError,
)
from undine.optimizer.optimizer import optimize_async, optimize_sync
from undine.optimizer.prefetch_hack import evaluate_with_prefetch_hack_async, evaluate_with_prefetch_hack_sync
from undine.relay import Node, from_global_id, offset_to_cursor, to_global_id
from undine.settings import undine_settings
from undine.typing import ConnectionDict, NodeDict, PageInfoDict, TModel
from undine.utils.graphql.undine_extensions import get_undine_query_type
from undine.utils.graphql.utils import (
    get_arguments,
    get_queried_field_name,
    get_underlying_type,
    pre_evaluate_request_user,
)
from undine.utils.model_utils import create_union_queryset
from undine.utils.reflection import get_root_and_info_params, is_subclass

if TYPE_CHECKING:
    from collections.abc import Callable, Hashable
    from types import FunctionType

    from django.db.models import Model, QuerySet
    from graphql.pyutils import AwaitableOrValue

    from undine import Entrypoint, Field, InterfaceType, UnionType
    from undine.dataclasses import FilterResults, OrderResults
    from undine.optimizer.optimizer import QueryOptimizer
    from undine.pagination import PaginationHandler
    from undine.relay import Connection
    from undine.typing import GQLInfo, QuerySetMap

__all__ = [
    "ConnectionResolver",
    "EntrypointFunctionResolver",
    "FieldFunctionResolver",
    "GlobalIDResolver",
    "InterfaceTypeResolver",
    "ModelAttributeResolver",
    "ModelManyRelatedFieldResolver",
    "ModelSingleRelatedFieldResolver",
    "NamedTupleFieldResolver",
    "NestedConnectionResolver",
    "NestedQueryTypeManyResolver",
    "NestedQueryTypeSingleResolver",
    "NodeResolver",
    "QueryTypeManyResolver",
    "QueryTypeSingleResolver",
    "TypedDictFieldResolver",
    "UnionTypeResolver",
]


@dataclasses.dataclass(frozen=True, slots=True)
class EntrypointFunctionResolver:
    """Resolves an `Entrypoint` using the given function."""

    func: FunctionType | Callable[..., Any]
    entrypoint: Entrypoint

    root_param: str | None = dataclasses.field(default=None, init=False)
    info_param: str | None = dataclasses.field(default=None, init=False)

    def __post_init__(self) -> None:
        params = get_root_and_info_params(self.func)
        object.__setattr__(self, "root_param", params.root_param)
        object.__setattr__(self, "info_param", params.info_param)

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        if undine_settings.ASYNC and inspect.iscoroutinefunction(self.func):
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        self.set_kwargs(kwargs, root, info)
        result = self.func(**kwargs)
        self.check_permissions(root, info, result)
        return result

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        self.set_kwargs(kwargs, root, info)
        result = await self.func(**kwargs)
        await self.check_permissions_async(root, info, result)
        return result

    def set_kwargs(self, kwargs: dict[str, Any], root: Any, info: GQLInfo) -> None:
        if self.root_param is not None:
            kwargs[self.root_param] = root
        if self.info_param is not None:
            kwargs[self.info_param] = info

    def check_permissions(self, root: Any, info: GQLInfo, result: Any) -> None:
        if self.entrypoint.permissions_func is not None:
            if self.entrypoint.many:
                for item in result:
                    self.entrypoint.permissions_func(root, info, item)
            else:
                self.entrypoint.permissions_func(root, info, result)

        elif is_subclass(self.entrypoint.ref, QueryType):
            self.entrypoint.ref.__permissions__(result, info)

    async def check_permissions_async(self, root: Any, info: GQLInfo, result: Any) -> None:
        if self.entrypoint.permissions_func is not None:
            if self.entrypoint.many:
                for item in result:
                    if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                        await self.entrypoint.permissions_func(root, info, item)
                    else:
                        self.entrypoint.permissions_func(root, info, item)

            elif inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                await self.entrypoint.permissions_func(root, info, result)

            else:
                self.entrypoint.permissions_func(root, info, result)

        elif is_subclass(self.entrypoint.ref, QueryType):
            if inspect.iscoroutinefunction(self.entrypoint.ref.__permissions__):
                await self.entrypoint.ref.__permissions__(result, info)
            else:
                self.entrypoint.ref.__permissions__(result, info)


@dataclasses.dataclass(frozen=True, slots=True)
class FieldFunctionResolver:
    """Resolves a `Field` using the given function."""

    func: FunctionType | Callable[..., Any]
    field: Field

    root_param: str | None = dataclasses.field(default=None, init=False)
    info_param: str | None = dataclasses.field(default=None, init=False)

    def __post_init__(self) -> None:
        params = get_root_and_info_params(self.func)
        object.__setattr__(self, "root_param", params.root_param)
        object.__setattr__(self, "info_param", params.info_param)

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        if undine_settings.ASYNC and inspect.iscoroutinefunction(self.func):
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        self.set_kwargs(kwargs, root, info)
        result = self.func(**kwargs)
        self.check_permissions(root, info, result)
        return result

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> Any:
        self.set_kwargs(kwargs, root, info)
        result = await self.func(**kwargs)
        await self.check_permissions_async(root, info, result)
        return result

    def set_kwargs(self, kwargs: dict[str, Any], root: Any, info: GQLInfo) -> None:
        if self.root_param is not None:
            kwargs[self.root_param] = root
        if self.info_param is not None:
            kwargs[self.info_param] = info

    def check_permissions(self, root: Any, info: GQLInfo, result: Any) -> None:
        if self.field.permissions_func is not None:
            if self.field.many:
                for item in result:
                    self.field.permissions_func(root, info, item)
            else:
                self.field.permissions_func(root, info, result)

        elif is_subclass(self.field.ref, QueryType):
            self.field.ref.__permissions__(result, info)

    async def check_permissions_async(self, root: Any, info: GQLInfo, result: Any) -> None:
        if self.field.permissions_func is not None:
            if self.field.many:
                for item in result:
                    if inspect.iscoroutinefunction(self.field.permissions_func):
                        await self.field.permissions_func(root, info, item)
                    else:
                        self.field.permissions_func(root, info, item)

            elif inspect.iscoroutinefunction(self.field.permissions_func):
                await self.field.permissions_func(root, info, result)

            else:
                self.field.permissions_func(root, info, result)

        elif is_subclass(self.field.ref, QueryType):
            if inspect.iscoroutinefunction(self.field.ref.__permissions__):
                await self.field.ref.__permissions__(result, info)
            else:
                self.field.ref.__permissions__(result, info)


# Model field resolvers


@dataclasses.dataclass(frozen=True, slots=True)
class ModelAttributeResolver:
    """Resolves a model field or annotation to a value by attribute access."""

    field: Field

    static: bool = True
    """
    If the attribute is queried multiple times in the same operation, should it return
    different values, for example, based on input arguments?
    """

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[Any]:
        if undine_settings.ASYNC:
            return self.run_async(root, info)
        return self.run_sync(root, info)

    def run_sync(self, root: Model, info: GQLInfo) -> Any:
        value = self.get_value(root, info)

        if value is None:
            if not self.field.nullable:
                raise GraphQLFieldNotNullableError(
                    typename=self.field.query_type.__schema_name__,
                    field_name=self.field.schema_name,
                )
            return None

        self.check_permissions(root, info, value)
        return value

    async def run_async(self, root: Model, info: GQLInfo) -> Any:
        value = self.get_value(root, info)

        if value is None:
            if not self.field.nullable:
                raise GraphQLFieldNotNullableError(
                    typename=self.field.query_type.__schema_name__,
                    field_name=self.field.schema_name,
                )
            return None

        await self.check_permissions_async(root, info, value)
        return value

    def get_value(self, root: Model, info: GQLInfo) -> Any:
        field_name = self.field.field_name
        if not self.static:
            field_name = get_queried_field_name(field_name, info)
        return getattr(root, field_name, None)

    def check_permissions(self, root: Model, info: GQLInfo, value: Any) -> None:
        if self.field.permissions_func is not None:
            self.field.permissions_func(root, info, value)

    async def check_permissions_async(self, root: Model, info: GQLInfo, value: Any) -> None:
        if self.field.permissions_func is not None:
            if inspect.iscoroutinefunction(self.field.permissions_func):
                await self.field.permissions_func(root, info, value)
            else:
                self.field.permissions_func(root, info, value)


@dataclasses.dataclass(frozen=True, slots=True)
class ModelSingleRelatedFieldResolver(Generic[TModel]):
    """Resolves single-related model field to its primary key."""

    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> Any:
        if undine_settings.ASYNC:
            return self.run_async(root, info)
        return self.run_sync(root, info)

    def run_sync(self, root: Model, info: GQLInfo) -> Any:
        value: TModel | None = getattr(root, self.field.field_name, None)
        if value is None:
            if not self.field.nullable:
                raise GraphQLFieldNotNullableError(
                    typename=self.field.query_type.__schema_name__,
                    field_name=self.field.schema_name,
                )
            return None

        self.check_permissions(root, info, value)
        return value.pk

    async def run_async(self, root: Model, info: GQLInfo) -> Any:
        value: TModel | None = getattr(root, self.field.field_name, None)
        if value is None:
            if not self.field.nullable:
                raise GraphQLFieldNotNullableError(
                    typename=self.field.query_type.__schema_name__,
                    field_name=self.field.schema_name,
                )
            return None

        await self.check_permissions_async(root, info, value)
        return value.pk

    def check_permissions(self, root: Model, info: GQLInfo, value: Any) -> None:
        if self.field.permissions_func is not None:
            self.field.permissions_func(root, info, value)

    async def check_permissions_async(self, root: Model, info: GQLInfo, value: Any) -> None:
        if self.field.permissions_func is not None:
            if inspect.iscoroutinefunction(self.field.permissions_func):
                await self.field.permissions_func(root, info, value)
            else:
                self.field.permissions_func(root, info, value)


@dataclasses.dataclass(frozen=True, slots=True)
class ModelManyRelatedFieldResolver(Generic[TModel]):
    """Resolves a many-related model field to a list of their primary keys."""

    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[Any]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info)
        return self.run_sync(root, info)

    def run_sync(self, root: Model, info: GQLInfo) -> list[Any]:
        instances = self.get_instances(root, info)
        self.check_permissions(root, info, instances)
        return [instance.pk for instance in instances]

    async def run_async(self, root: Model, info: GQLInfo) -> list[Any]:
        instances = self.get_instances(root, info)
        await self.check_permissions_async(root, info, instances)
        return [instance.pk for instance in instances]

    def get_instances(self, root: Model, info: GQLInfo) -> list[TModel]:
        field_name = get_queried_field_name(self.field.field_name, info)
        manager: BaseManager[TModel] = getattr(root, field_name)
        return list(manager.get_queryset())

    def check_permissions(self, root: Model, info: GQLInfo, instances: list[TModel]) -> None:
        if self.field.permissions_func is not None:
            for instance in instances:
                self.field.permissions_func(root, info, instance)

    async def check_permissions_async(self, root: Model, info: GQLInfo, instances: list[TModel]) -> None:
        if self.field.permissions_func is not None:
            for instance in instances:
                if inspect.iscoroutinefunction(self.field.permissions_func):
                    await self.field.permissions_func(root, info, instance)
                else:
                    self.field.permissions_func(root, info, instance)


@dataclasses.dataclass(frozen=True, slots=True)
class ModelGenericForeignKeyResolver(Generic[TModel]):
    """Resolves generic foreign key field to its related model instance."""

    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[TModel | None]:
        if undine_settings.ASYNC:
            return self.run_async(root, info)
        return self.run_sync(root, info)

    def run_sync(self, root: Model, info: GQLInfo) -> TModel | None:
        value: TModel | None = getattr(root, self.field.field_name, None)
        if value is None:
            return None

        self.check_permissions(root, info, value)
        return value

    async def run_async(self, root: Model, info: GQLInfo) -> TModel | None:
        value: TModel | None = getattr(root, self.field.field_name, None)
        if value is None:
            return None

        await self.check_permissions_async(root, info, value)
        return value

    def check_permissions(self, root: Model, info: GQLInfo, value: Any) -> None:
        if self.field.permissions_func is not None:
            self.field.permissions_func(root, info, value)

    async def check_permissions_async(self, root: Model, info: GQLInfo, value: Any) -> None:
        if self.field.permissions_func is not None:
            if inspect.iscoroutinefunction(self.field.permissions_func):
                await self.field.permissions_func(root, info, value)
            else:
                self.field.permissions_func(root, info, value)


# Query type resolvers


@dataclasses.dataclass(frozen=True, slots=True)
class QueryTypeSingleResolver(Generic[TModel]):
    """Top-level resolver for fetching a single model object via a QueryType."""

    query_type: type[QueryType[TModel]]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[TModel | None]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        queryset = self.query_type.__get_queryset__(info)
        instance = optimize_sync(queryset, info, **kwargs)

        if instance is None:
            if not self.entrypoint.nullable:
                raise GraphQLModelNotFoundError(model=self.query_type, pk=kwargs.get("pk"))
            return None

        self.check_permissions(root, info, instance)
        return instance

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        queryset = self.query_type.__get_queryset__(info)
        instance = await optimize_async(queryset, info, **kwargs)

        if instance is None:
            if not self.entrypoint.nullable:
                raise GraphQLModelNotFoundError(model=self.query_type, pk=kwargs.get("pk"))
            return None

        await self.check_permissions_async(root, info, instance)
        return instance

    def check_permissions(self, root: Any, info: GQLInfo, instance: TModel) -> None:
        if self.entrypoint.permissions_func is not None:
            self.entrypoint.permissions_func(root, info, instance)
        else:
            self.query_type.__permissions__(instance, info)

    async def check_permissions_async(self, root: Any, info: GQLInfo, instance: TModel) -> None:
        if self.entrypoint.permissions_func is not None:
            if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                await self.entrypoint.permissions_func(root, info, instance)
            else:
                self.entrypoint.permissions_func(root, info, instance)

        elif inspect.iscoroutinefunction(self.query_type.__permissions__):
            await self.query_type.__permissions__(instance, info)

        else:
            self.query_type.__permissions__(instance, info)


@dataclasses.dataclass(frozen=True, slots=True)
class QueryTypeManyResolver(Generic[TModel]):
    """Top-level resolver for fetching a set of model objects via a QueryType."""

    query_type: type[QueryType[TModel]]
    entrypoint: Entrypoint

    additional_filter: Optional[Q] = None  # noqa: UP045

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        queryset = self.get_queryset(info)
        instances = optimize_sync(queryset, info, limit=self.entrypoint.limit)
        self.check_permissions(root, info, instances)
        return instances

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        queryset = self.get_queryset(info)
        instances = await optimize_async(queryset, info, limit=self.entrypoint.limit)
        await self.check_permissions_async(root, info, instances)
        return instances

    def get_queryset(self, info: GQLInfo) -> QuerySet[TModel]:
        queryset = self.query_type.__get_queryset__(info)
        if self.additional_filter is not None:
            queryset = queryset.filter(self.additional_filter)
        return queryset

    def check_permissions(self, root: Any, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                self.entrypoint.permissions_func(root, info, instance)
            else:
                self.query_type.__permissions__(instance, info)

    async def check_permissions_async(self, root: Any, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                    await self.entrypoint.permissions_func(root, info, instance)
                else:
                    self.entrypoint.permissions_func(root, info, instance)

            elif inspect.iscoroutinefunction(self.query_type.__permissions__):
                await self.query_type.__permissions__(instance, info)

            else:
                self.query_type.__permissions__(instance, info)


@dataclasses.dataclass(frozen=True, slots=True)
class NestedQueryTypeSingleResolver(Generic[TModel]):
    """Resolves a single-related field pointing to another QueryType."""

    query_type: type[QueryType[TModel]]
    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[TModel | None]:
        if undine_settings.ASYNC:
            return self.run_async(root, info)
        return self.run_sync(root, info)

    def run_sync(self, root: Model, info: GQLInfo) -> TModel | None:
        instance: TModel | None = getattr(root, self.field.field_name, None)

        if instance is None:
            if not self.field.nullable:
                raise GraphQLFieldNotNullableError(
                    typename=self.query_type.__schema_name__,
                    field_name=self.field.schema_name,
                )
            return None

        self.check_permissions(root, info, instance)
        return instance

    async def run_async(self, root: Model, info: GQLInfo) -> TModel | None:
        instance: TModel | None = getattr(root, self.field.field_name, None)

        if instance is None:
            if not self.field.nullable:
                raise GraphQLFieldNotNullableError(
                    typename=self.query_type.__schema_name__,
                    field_name=self.field.schema_name,
                )
            return None

        await self.check_permissions_async(root, info, instance)
        return instance

    def check_permissions(self, root: Any, info: GQLInfo, instance: TModel) -> None:
        if self.field.permissions_func is not None:
            self.field.permissions_func(root, info, instance)
        else:
            self.query_type.__permissions__(instance, info)

    async def check_permissions_async(self, root: Any, info: GQLInfo, instance: TModel) -> None:
        if self.field.permissions_func is not None:
            if inspect.iscoroutinefunction(self.field.permissions_func):
                await self.field.permissions_func(root, info, instance)
            else:
                self.field.permissions_func(root, info, instance)

        elif inspect.iscoroutinefunction(self.query_type.__permissions__):
            await self.query_type.__permissions__(instance, info)

        else:
            self.query_type.__permissions__(instance, info)


@dataclasses.dataclass(frozen=True, slots=True)
class NestedQueryTypeManyResolver(Generic[TModel]):
    """Resolves a many-related field pointing to another QueryType."""

    query_type: type[QueryType[TModel]]
    field: Field

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info)
        return self.run_sync(root, info)

    def run_sync(self, root: Model, info: GQLInfo) -> list[TModel]:
        instances = self.get_instances(root, info)
        self.check_permissions(root, info, instances)
        return instances

    async def run_async(self, root: Model, info: GQLInfo) -> list[TModel]:
        instances = self.get_instances(root, info)
        await self.check_permissions_async(root, info, instances)
        return instances

    def get_instances(self, root: Model, info: GQLInfo) -> list[TModel]:
        field_name = get_queried_field_name(self.field.field_name, info)

        instances: list[TModel] = getattr(root, field_name)
        if isinstance(instances, BaseManager):
            instances = list(instances.get_queryset())
        return instances

    def check_permissions(self, root: Model, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.field.permissions_func is not None:
                self.field.permissions_func(root, info, instance)
            else:
                self.query_type.__permissions__(instance, info)

    async def check_permissions_async(self, root: Model, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.field.permissions_func is not None:
                if inspect.iscoroutinefunction(self.field.permissions_func):
                    await self.field.permissions_func(root, info, instance)
                else:
                    self.field.permissions_func(root, info, instance)

            elif inspect.iscoroutinefunction(self.query_type.__permissions__):
                await self.query_type.__permissions__(instance, info)

            else:
                self.query_type.__permissions__(instance, info)


# Relay


@dataclasses.dataclass(frozen=True, slots=True)
class GlobalIDResolver:
    """Resolves a model primary key as a Global ID."""

    typename: str

    def __call__(self, root: Model, info: GQLInfo, **kwargs: Any) -> str:
        return to_global_id(self.typename, root.pk)


@dataclasses.dataclass(frozen=True, slots=True)
class NodeResolver(Generic[TModel]):
    """Resolves a model instance through a Global ID."""

    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[TModel | None]:
        try:
            typename, object_id = from_global_id(kwargs["id"])
        except Exception as error:
            raise GraphQLNodeInvalidGlobalIDError(value=kwargs["id"]) from error

        object_type = info.schema.get_type(typename)
        if object_type is None:
            raise GraphQLNodeObjectTypeMissingError(typename=typename)

        if not isinstance(object_type, GraphQLObjectType):
            raise GraphQLNodeTypeNotObjectTypeError(typename=typename)

        query_type = get_undine_query_type(object_type)
        if query_type is None:
            raise GraphQLNodeQueryTypeMissingError(typename=typename)

        if Node not in query_type.__interfaces__:
            raise GraphQLNodeInterfaceMissingError(typename=typename)

        field: Field | None = query_type.__field_map__.get("id")
        if field is None:
            raise GraphQLNodeMissingIDFieldError(typename=typename)

        field_type = get_underlying_type(field.get_field_type())  # type: ignore[type-var]
        if field_type is not GraphQLID:
            raise GraphQLNodeIDFieldTypeError(typename=typename)

        resolver = QueryTypeSingleResolver(query_type=query_type, entrypoint=self.entrypoint)
        return resolver(root, info, pk=object_id)


@dataclasses.dataclass(frozen=True, slots=True)
class ConnectionResolver(Generic[TModel]):
    """Resolves a connection of items."""

    connection: Connection
    entrypoint: Entrypoint

    @property
    def query_type(self) -> type[QueryType]:
        return self.connection.query_type  # type: ignore[return-value]

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[ConnectionDict[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info)
        return self.run_sync(root, info)

    def run_sync(self, root: Any, info: GQLInfo) -> ConnectionDict[TModel]:
        results = self.run_optimizer(info)
        instances = evaluate_with_prefetch_hack_sync(results.queryset)
        self.check_permissions(root, info, instances)
        return self.to_connection(instances, pagination=results.pagination)

    async def run_async(self, root: Any, info: GQLInfo) -> ConnectionDict[TModel]:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        results = await self.run_optimizer_async(info)
        instances = await evaluate_with_prefetch_hack_async(results.queryset)
        await self.check_permissions_async(root, info, instances)
        return self.to_connection(instances, pagination=results.pagination)

    def get_queryset(self, info: GQLInfo) -> QuerySet[TModel]:
        return self.query_type.__get_queryset__(info)

    def run_optimizer(self, info: GQLInfo) -> OptimizationWithPagination[TModel]:
        queryset = self.get_queryset(info)
        optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=queryset.model, info=info)
        optimizations = optimizer.compile()
        optimized_queryset = optimizations.apply(queryset, info)
        return OptimizationWithPagination(
            queryset=optimized_queryset,
            pagination=optimizations.pagination,  # type: ignore[arg-type]
        )

    async def run_optimizer_async(self, info: GQLInfo) -> OptimizationWithPagination[TModel]:
        queryset = self.get_queryset(info)
        optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=queryset.model, info=info)
        optimizations = optimizer.compile()
        # Applying may call 'queryset.count()'.
        optimized_queryset = await sync_to_async(optimizations.apply)(queryset, info)
        return OptimizationWithPagination(
            queryset=optimized_queryset,
            pagination=optimizations.pagination,  # type: ignore[arg-type]
        )

    def check_permissions(self, root: Any, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                self.entrypoint.permissions_func(root, info, instance)
            else:
                self.query_type.__permissions__(instance, info)

    async def check_permissions_async(self, root: Any, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                    await self.entrypoint.permissions_func(root, info, instance)
                else:
                    self.entrypoint.permissions_func(root, info, instance)

            elif inspect.iscoroutinefunction(self.query_type.__permissions__):
                await self.query_type.__permissions__(instance, info)

            else:
                self.query_type.__permissions__(instance, info)

    def to_connection(self, instances: list[TModel], pagination: PaginationHandler) -> ConnectionDict[TModel]:
        typename = self.query_type.__schema_name__
        edges = [
            NodeDict(
                cursor=offset_to_cursor(typename, pagination.start + index),
                node=instance,
            )
            for index, instance in enumerate(instances)
        ]
        return ConnectionDict(
            totalCount=pagination.total_count or 0,
            pageInfo=PageInfoDict(
                hasNextPage=(
                    False
                    if pagination.stop is None
                    else True
                    if pagination.total_count is None
                    else pagination.stop < pagination.total_count
                ),
                hasPreviousPage=pagination.start > 0,
                startCursor=None if not edges else edges[0]["cursor"],
                endCursor=None if not edges else edges[-1]["cursor"],
            ),
            edges=edges,
        )


@dataclasses.dataclass(frozen=True, slots=True)
class NestedConnectionResolver(Generic[TModel]):
    """Resolves a nested connection from the given field."""

    connection: Connection
    field: Field

    @property
    def query_type(self) -> type[QueryType]:
        return self.connection.query_type  # type: ignore[return-value]

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[ConnectionDict[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info)
        return self.run_sync(root, info)

    def run_sync(self, root: Model, info: GQLInfo, **kwargs: Any) -> ConnectionDict[TModel]:
        field_name = get_queried_field_name(self.field.field_name, info)
        instances = self.get_instances(root, field_name)
        self.check_permissions(root, info, instances)
        return self.to_connection(instances)

    async def run_async(self, root: Model, info: GQLInfo, **kwargs: Any) -> ConnectionDict[TModel]:
        field_name = get_queried_field_name(self.field.field_name, info)
        instances = self.get_instances(root, field_name)
        await self.check_permissions_async(root, info, instances)
        return self.to_connection(instances)

    def get_instances(self, root: Model, field_name: str) -> list[TModel]:
        instances: list[TModel] = getattr(root, field_name)
        if isinstance(instances, BaseManager):
            instances = list(instances.get_queryset())
        return instances

    def check_permissions(self, root: Any, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.field.permissions_func is not None:
                self.field.permissions_func(root, info, instance)
            else:
                self.query_type.__permissions__(instance, info)

    async def check_permissions_async(self, root: Any, info: GQLInfo, instances: list[TModel]) -> None:
        for instance in instances:
            if self.field.permissions_func is not None:
                if inspect.iscoroutinefunction(self.field.permissions_func):
                    await self.field.permissions_func(root, info, instance)
                else:
                    self.field.permissions_func(root, info, instance)

            elif inspect.iscoroutinefunction(self.query_type.__permissions__):
                await self.query_type.__permissions__(instance, info)

            else:
                self.query_type.__permissions__(instance, info)

    def get_pagination_params(self, instances: list[TModel]) -> tuple[int, int | None, int | None]:
        total_count: int | None = None
        start: int = 0
        stop: int | None = None

        # Not optimal, as we don't know the actual pagination params if there are no results.
        if instances:
            total_count = getattr(instances[0], undine_settings.PAGINATION_TOTAL_COUNT_KEY, None)
            start = getattr(instances[0], undine_settings.PAGINATION_START_INDEX_KEY, 0)
            stop = getattr(instances[0], undine_settings.PAGINATION_STOP_INDEX_KEY, None)

        return start, stop, total_count

    def to_connection(self, instances: list[TModel]) -> ConnectionDict[TModel]:
        start, stop, total_count = self.get_pagination_params(instances)

        typename = self.query_type.__schema_name__
        edges = [
            NodeDict(
                cursor=offset_to_cursor(typename, start + index),
                node=instance,
            )
            for index, instance in enumerate(instances)
        ]
        return ConnectionDict(
            totalCount=total_count or 0,
            pageInfo=PageInfoDict(
                hasNextPage=(False if stop is None else True if total_count is None else stop < total_count),
                hasPreviousPage=start > 0,
                startCursor=None if not edges else edges[0]["cursor"],
                endCursor=None if not edges else edges[-1]["cursor"],
            ),
            edges=edges,
        )


# UnionType


@dataclasses.dataclass(frozen=True, slots=True)
class UnionTypeResolver(Generic[TModel]):
    """Top-level resolver for fetching a set of model objects from a `UnionType`."""

    union_type: type[UnionType]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        queryset_map = self.optimize(info, **kwargs)
        return self.fetch_instances(root, info, queryset_map)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        queryset_map = self.optimize(info, **kwargs)
        return await self.fetch_instances_async(root, info, queryset_map)

    def fetch_instances(self, root: Any, info: GQLInfo, queryset_map: QuerySetMap) -> list[TModel]:
        """
        Fetch all instances from the querysets.

        Use a union query to filter and order the results together before fetching them.
        Then fetch the instances per-model and sort them in the same order as in the union query.
        """
        arg_values = get_arguments(info)

        union_qs = create_union_queryset(queryset_map.values())
        union_qs = union_qs.order_by("pk")  # Default ordering

        if self.union_type.__filterset__:
            filter_results = self.filter_union(arg_values, info, queryset_map)
            if filter_results.none:
                return []

        if self.union_type.__orderset__:
            order_results = self.order_union(arg_values, info, queryset_map)
            union_qs = union_qs.order_by(*order_results.order_by)

        union_qs = union_qs.values("__typename", "pk")

        if self.entrypoint.limit is not None:
            union_qs = union_qs[: self.entrypoint.limit]

        # Store the order in which the union query returned the items
        # from each model, so that we can sort when the actual instances are fetched.
        pks_by_typename: dict[str, dict[str, int]] = defaultdict(dict)
        for index, item in enumerate(union_qs):
            pk = self.coalesce_pk(item["pk"])
            pks_by_typename[item["__typename"]][pk] = index

        all_instances: dict[int, TModel] = {}
        query_type_by_name = {query_type.__schema_name__: query_type for query_type in queryset_map}

        # Fetch all instances for each model, if some primary keys for it were returned in the union queryset.
        for typename, pk_to_index in pks_by_typename.items():
            query_type = query_type_by_name[typename]
            queryset = queryset_map[query_type]
            queryset = queryset.filter(pk__in=list(pk_to_index))
            instances = evaluate_with_prefetch_hack_sync(queryset)
            self.check_permissions(root, info, query_type, instances)

            # Set the instances in the same order as the union query returned them.
            for instance in instances:
                pk = self.coalesce_pk(instance.pk)
                all_instances[pk_to_index[pk]] = instance

        return [instance for _, instance in sorted(all_instances.items())]

    async def fetch_instances_async(self, root: Any, info: GQLInfo, queryset_map: QuerySetMap) -> list[TModel]:
        arg_values = get_arguments(info)

        union_qs = create_union_queryset(queryset_map.values())
        union_qs = union_qs.order_by("pk")  # Default ordering

        if self.union_type.__filterset__:
            filter_results = self.filter_union(arg_values, info, queryset_map)
            if filter_results.none:
                return []

        if self.union_type.__orderset__:
            order_results = self.order_union(arg_values, info, queryset_map)
            union_qs = union_qs.order_by(*order_results.order_by)

        union_qs = union_qs.values("__typename", "pk")

        if self.entrypoint.limit is not None:
            union_qs = union_qs[: self.entrypoint.limit]

        # Store the order in which the union query returned the items
        # from each model, so that we can sort when the actual instances are fetched.
        pks_by_typename: dict[str, dict[str, Any]] = defaultdict(dict)
        counter = count()
        async for item in union_qs:
            pk = self.coalesce_pk(item["pk"])
            pks_by_typename[item["__typename"]][pk] = next(counter)

        all_instances: dict[int, TModel] = {}
        query_type_by_name = {query_type.__schema_name__: query_type for query_type in queryset_map}

        # Fetch all instances for each model, if some primary keys for it were returned in the union queryset.
        for typename, pk_to_index in pks_by_typename.items():
            query_type = query_type_by_name[typename]
            queryset = queryset_map[query_type]
            queryset = queryset.filter(pk__in=list(pk_to_index))
            instances = await evaluate_with_prefetch_hack_async(queryset)
            await self.check_permissions_async(root, info, query_type, instances)

            # Set the instances in the same order as the union query returned them.
            for instance in instances:
                pk = self.coalesce_pk(instance.pk)
                all_instances[pk_to_index[pk]] = instance

        return [instance for _, instance in sorted(all_instances.items())]

    def check_permissions(
        self,
        root: Any,
        info: GQLInfo,
        query_type: type[QueryType[TModel]],
        instances: list[TModel],
    ) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                self.entrypoint.permissions_func(root, info, instance)
            else:
                query_type.__permissions__(instance, info)

    async def check_permissions_async(
        self,
        root: Any,
        info: GQLInfo,
        query_type: type[QueryType[TModel]],
        instances: list[TModel],
    ) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                    await self.entrypoint.permissions_func(root, info, instance)
                else:
                    self.entrypoint.permissions_func(root, info, instance)

            elif inspect.iscoroutinefunction(query_type.__permissions__):
                await query_type.__permissions__(instance, info)

            else:
                query_type.__permissions__(instance, info)

    def filter_union(
        self,
        arg_values: dict[str, Any],
        info: GQLInfo,
        queryset_map: dict[type[QueryType], QuerySet[Any, Any]],
    ) -> FilterResults:
        filter_data = arg_values.get(undine_settings.QUERY_TYPE_FILTER_INPUT_KEY, {})
        filter_results = self.union_type.__filterset__.__build__(filter_data, info)
        if filter_results.none:
            return filter_results

        for query_type, queryset in queryset_map.items():
            if filter_results.aliases:
                queryset = queryset.alias(**filter_results.aliases)  # noqa: PLW2901
            if filter_results.distinct:
                queryset = queryset.distinct()  # noqa: PLW2901
            if filter_results.filters:
                queryset = queryset.filter(Q(*filter_results.filters))  # noqa: PLW2901

            queryset_map[query_type] = queryset

        return filter_results

    def order_union(
        self,
        arg_values: dict[str, Any],
        info: GQLInfo,
        queryset_map: dict[type[QueryType], QuerySet[Any, Any]],
    ) -> OrderResults:
        order_data = arg_values.get(undine_settings.QUERY_TYPE_ORDER_INPUT_KEY, [])
        order_results = self.union_type.__orderset__.__build__(order_data, info)

        for query_type, queryset in queryset_map.items():
            if order_results.aliases:
                queryset = queryset.alias(**order_results.aliases)  # noqa: PLW2901
            if order_results.order_by:
                queryset = queryset.order_by(*order_results.order_by, *queryset.query.order_by)  # noqa: PLW2901

            queryset_map[query_type] = queryset

        return order_results

    def optimize(self, info: GQLInfo, **kwargs: Any) -> QuerySetMap:
        queryset_map: QuerySetMap = {}

        for model, query_type in self.union_type.__query_types_by_model__.items():
            optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=model, info=info)
            optimizations = optimizer.compile()

            # If nothing is selected from this union type member, don't fetch anything from it
            if not optimizations:
                continue

            args: dict[str, Any] = {}
            filter_key = f"{undine_settings.QUERY_TYPE_FILTER_INPUT_KEY}{model.__name__}"
            order_by_key = f"{undine_settings.QUERY_TYPE_ORDER_INPUT_KEY}{model.__name__}"

            if filter_key in kwargs:
                args[undine_settings.QUERY_TYPE_FILTER_INPUT_KEY] = kwargs[filter_key]
            if order_by_key in kwargs:
                args[undine_settings.QUERY_TYPE_ORDER_INPUT_KEY] = kwargs[order_by_key]

            optimizer.handle_undine_query_type(query_type, args)

            optimizations.annotations["__typename"] = Value(query_type.__schema_name__)

            queryset = query_type.__get_queryset__(info)
            queryset_map[query_type] = optimizations.apply(queryset, info)

        return queryset_map

    def coalesce_pk(self, item: Any) -> Any:
        # Converting UUIDs to hex avoids an issue with union querysets with mixed int and UUID pks.
        return item.hex if isinstance(item, uuid.UUID) else item


@dataclasses.dataclass(frozen=True, slots=True)
class UnionTypeConnectionResolver(Generic[TModel]):
    """Resolves a connection of union type items."""

    connection: Connection
    entrypoint: Entrypoint

    @property
    def union_type(self) -> type[UnionType]:
        return self.connection.union_type  # type: ignore[return-value]

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[ConnectionDict[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> ConnectionDict[TModel]:
        result = self.optimize(info, **kwargs)
        if not result.queryset_map:
            return self.empty_connection()

        all_instances = self.fetch_instances(root, info, result)
        return self.to_connection(all_instances, pagination=result.pagination)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> ConnectionDict[TModel]:
        result = self.optimize(info, **kwargs)
        if not result.queryset_map:
            return self.empty_connection()

        all_instances = await self.fetch_instances_async(root, info, result)
        return self.to_connection(all_instances, pagination=result.pagination)

    def fetch_instances(self, root: Any, info: GQLInfo, result: QuerySetMapWithPagination) -> list[TModel]:
        """
        Fetch paginated instances from the querysets.

        Use a union query to filter, order, and paginate the results together before fetching them.
        Then fetch the instances per-model and sort them in the same order as in the union query.
        """
        arg_values = get_arguments(info)
        queryset_map = result.queryset_map
        pagination = result.pagination

        union_qs = create_union_queryset(queryset_map.values())
        union_qs = union_qs.order_by("pk")  # Default ordering

        if self.union_type.__filterset__:
            filter_results = self.filter_union(arg_values, info, queryset_map)
            if filter_results.none:
                return []

        if self.union_type.__orderset__:
            order_results = self.order_union(arg_values, info, queryset_map)
            union_qs = union_qs.order_by(*order_results.order_by)

        union_qs = union_qs.values("__typename", "pk")
        union_qs = pagination.paginate_queryset(union_qs, info)

        # Store the order in which the union query returned the items
        # from each model, so that we can sort when the actual instances are fetched.
        pks_by_typename: dict[str, dict[Hashable, int]] = defaultdict(dict)
        for index, item in enumerate(union_qs):
            pk = self.coalesce_pk(item["pk"])
            pks_by_typename[item["__typename"]][pk] = index

        all_instances: dict[int, TModel] = {}
        query_type_by_name = {query_type.__schema_name__: query_type for query_type in queryset_map}

        # Fetch all instances for each model, if some primary keys for it were returned in the union queryset.
        for typename, pk_to_index in pks_by_typename.items():
            query_type = query_type_by_name[typename]
            queryset = queryset_map[query_type]
            queryset = queryset.filter(pk__in=list(pk_to_index))
            instances = evaluate_with_prefetch_hack_sync(queryset)
            self.check_permissions(root, info, query_type, instances)

            # Set the instances in the same order as the union query returned them.
            for instance in instances:
                pk = self.coalesce_pk(instance.pk)
                all_instances[pk_to_index[pk]] = instance

        return [instance for _, instance in sorted(all_instances.items())]

    async def fetch_instances_async(self, root: Any, info: GQLInfo, result: QuerySetMapWithPagination) -> list[TModel]:
        """Async version of `fetch_instances`."""
        arg_values = get_arguments(info)

        queryset_map = result.queryset_map
        pagination = result.pagination

        union_qs = create_union_queryset(queryset_map.values())
        union_qs = union_qs.order_by("pk")

        if self.union_type.__filterset__:
            filter_results = self.filter_union(arg_values, info, queryset_map)
            if filter_results.none:
                return []

        if self.union_type.__orderset__:
            order_results = self.order_union(arg_values, info, queryset_map)
            union_qs = union_qs.order_by(*order_results.order_by)

        union_qs = union_qs.values("__typename", "pk")
        # May call 'queryset.count()'.
        union_qs = await sync_to_async(pagination.paginate_queryset)(union_qs, info)

        # Store the order in which the union query returned the items
        # from each model, so that we can sort when the actual instances are fetched.
        pks_by_typename: dict[str, dict[Hashable, int]] = defaultdict(dict)
        counter = count()
        async for item in union_qs:
            pk = self.coalesce_pk(item["pk"])
            pks_by_typename[item["__typename"]][pk] = next(counter)

        all_instances: dict[int, TModel] = {}
        query_type_by_name = {query_type.__schema_name__: query_type for query_type in queryset_map}

        # Fetch all instances for each model, if some primary keys for it were returned in the union queryset.
        for typename, pk_to_index in pks_by_typename.items():
            query_type = query_type_by_name[typename]
            queryset = queryset_map[query_type]
            queryset = queryset.filter(pk__in=list(pk_to_index))
            instances = await evaluate_with_prefetch_hack_async(queryset)
            await self.check_permissions_async(root, info, query_type, instances)

            # Set the instances in the same order as the union query returned them.
            for instance in instances:
                pk = self.coalesce_pk(instance.pk)
                all_instances[pk_to_index[pk]] = instance

        return [instance for _, instance in sorted(all_instances.items())]

    def check_permissions(
        self,
        root: Any,
        info: GQLInfo,
        query_type: type[QueryType[TModel]],
        instances: list[TModel],
    ) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                self.entrypoint.permissions_func(root, info, instance)
            else:
                query_type.__permissions__(instance, info)

    async def check_permissions_async(
        self,
        root: Any,
        info: GQLInfo,
        query_type: type[QueryType[TModel]],
        instances: list[TModel],
    ) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                    await self.entrypoint.permissions_func(root, info, instance)
                else:
                    self.entrypoint.permissions_func(root, info, instance)

            elif inspect.iscoroutinefunction(query_type.__permissions__):
                await query_type.__permissions__(instance, info)

            else:
                query_type.__permissions__(instance, info)

    def filter_union(
        self,
        arg_values: dict[str, Any],
        info: GQLInfo,
        queryset_map: dict[type[QueryType], QuerySet[Any, Any]],
    ) -> FilterResults:
        filter_data = arg_values.get(undine_settings.QUERY_TYPE_FILTER_INPUT_KEY, {})
        filter_results = self.union_type.__filterset__.__build__(filter_data, info)
        if filter_results.none:
            return filter_results

        for query_type, queryset in queryset_map.items():
            if filter_results.aliases:
                queryset = queryset.alias(**filter_results.aliases)  # noqa: PLW2901
            if filter_results.distinct:
                queryset = queryset.distinct()  # noqa: PLW2901
            if filter_results.filters:
                queryset = queryset.filter(Q(*filter_results.filters))  # noqa: PLW2901

            queryset_map[query_type] = queryset

        return filter_results

    def order_union(
        self,
        arg_values: dict[str, Any],
        info: GQLInfo,
        queryset_map: dict[type[QueryType], QuerySet[Any, Any]],
    ) -> OrderResults:
        order_data = arg_values.get(undine_settings.QUERY_TYPE_ORDER_INPUT_KEY, [])
        order_results = self.union_type.__orderset__.__build__(order_data, info)

        for query_type, queryset in queryset_map.items():
            if order_results.aliases:
                queryset = queryset.alias(**order_results.aliases)  # noqa: PLW2901
            if order_results.order_by:
                queryset = queryset.order_by(*order_results.order_by, *queryset.query.order_by)  # noqa: PLW2901

            queryset_map[query_type] = queryset

        return order_results

    def optimize(self, info: GQLInfo, **kwargs: Any) -> QuerySetMapWithPagination[TModel]:
        queryset_map: QuerySetMap = {}
        pagination: PaginationHandler | None = None

        for model, query_type in self.union_type.__query_types_by_model__.items():
            optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=model, info=info)
            optimizations = optimizer.compile()

            # If nothing is selected from this union type member, don't fetch anything from it
            if not optimizations:
                continue

            args: dict[str, Any] = {}
            filter_key = f"{undine_settings.QUERY_TYPE_FILTER_INPUT_KEY}{model.__name__}"
            order_by_key = f"{undine_settings.QUERY_TYPE_ORDER_INPUT_KEY}{model.__name__}"

            if filter_key in kwargs:
                args[undine_settings.QUERY_TYPE_FILTER_INPUT_KEY] = kwargs[filter_key]
            if order_by_key in kwargs:
                args[undine_settings.QUERY_TYPE_ORDER_INPUT_KEY] = kwargs[order_by_key]

            optimizer.handle_undine_query_type(query_type, args)

            optimizations.annotations["__typename"] = Value(query_type.__schema_name__)

            # Save pagination (should be the same for all query types)
            pagination = optimizations.pagination
            optimizations.pagination = None

            queryset = query_type.__get_queryset__(info)
            queryset_map[query_type] = optimizations.apply(queryset, info)

        return QuerySetMapWithPagination(queryset_map=queryset_map, pagination=pagination)

    def to_connection(self, instances: list[TModel], pagination: PaginationHandler) -> ConnectionDict[TModel]:
        typename = self.union_type.__schema_name__
        edges = [
            NodeDict(
                cursor=offset_to_cursor(typename, pagination.start + index),
                node=instance,
            )
            for index, instance in enumerate(instances)
        ]
        return ConnectionDict(
            totalCount=pagination.total_count or 0,
            pageInfo=PageInfoDict(
                hasNextPage=(
                    False
                    if pagination.stop is None
                    else True
                    if pagination.total_count is None
                    else pagination.stop < pagination.total_count
                ),
                hasPreviousPage=pagination.start > 0,
                startCursor=None if not edges else edges[0]["cursor"],
                endCursor=None if not edges else edges[-1]["cursor"],
            ),
            edges=edges,
        )

    def empty_connection(self) -> ConnectionDict[TModel]:
        page_info = PageInfoDict(hasNextPage=False, hasPreviousPage=False, startCursor=None, endCursor=None)
        return ConnectionDict(totalCount=0, pageInfo=page_info, edges=[])

    def coalesce_pk(self, item: Any) -> Any:
        # Converting UUIDs to hex avoids an issue with union querysets with mixed int and UUID pks.
        return item.hex if isinstance(item, uuid.UUID) else item


# InterfaceType


@dataclasses.dataclass(frozen=True, slots=True)
class InterfaceTypeResolver(Generic[TModel]):
    """Resolves an interface type to all of its implementations."""

    interface: type[InterfaceType]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_sync_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        queryset_map = self.optimize(info, **kwargs)
        return self.fetch_instances(info, root, queryset_map)

    async def run_sync_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        queryset_map = self.optimize(info, **kwargs)
        return await self.fetch_instances_async(info, root, queryset_map)

    def fetch_instances(self, info: GQLInfo, root: Any, queryset_map: QuerySetMap) -> list[TModel]:
        union_qs = create_union_queryset(queryset_map.values())
        union_qs = union_qs.order_by("pk")  # Default ordering

        # TODO: Filtering and ordering

        union_qs = union_qs.values("__typename", "pk")

        if self.entrypoint.limit is not None:
            union_qs = union_qs[: self.entrypoint.limit]

        # Store the order in which the union query returned the items
        # from each model, so that we can sort when the actual instances are fetched.
        pks_by_typename: dict[str, dict[Hashable, int]] = defaultdict(dict)
        for index, item in enumerate(union_qs):
            pk = self.coalesce_pk(item["pk"])
            pks_by_typename[item["__typename"]][pk] = index

        all_instances: dict[int, TModel] = {}
        query_type_by_name = {query_type.__schema_name__: query_type for query_type in queryset_map}

        # Fetch all instances for each model, if some primary keys for it were returned in the union queryset.
        for typename, pk_to_index in pks_by_typename.items():
            query_type = query_type_by_name[typename]
            queryset = queryset_map[query_type]
            queryset = queryset.filter(pk__in=list(pk_to_index))
            instances = evaluate_with_prefetch_hack_sync(queryset)
            self.check_permissions(root, info, query_type, instances)

            # Set the instances in the same order as the union query returned them.
            for instance in instances:
                pk = self.coalesce_pk(instance.pk)
                all_instances[pk_to_index[pk]] = instance

        return [instance for _, instance in sorted(all_instances.items())]

    async def fetch_instances_async(self, info: GQLInfo, root: Any, queryset_map: QuerySetMap) -> list[TModel]:
        union_qs = create_union_queryset(queryset_map.values())
        union_qs = union_qs.order_by("pk")  # Default ordering

        # TODO: Filtering and ordering

        union_qs = union_qs.values("__typename", "pk")

        if self.entrypoint.limit is not None:
            union_qs = union_qs[: self.entrypoint.limit]

        # Store the order in which the union query returned the items
        # from each model, so that we can sort when the actual instances are fetched.
        pks_by_typename: dict[str, dict[Hashable, Any]] = defaultdict(dict)
        counter = count()
        async for item in union_qs:
            pk = self.coalesce_pk(item["pk"])
            pks_by_typename[item["__typename"]][pk] = next(counter)

        all_instances: dict[int, TModel] = {}
        query_type_by_name = {query_type.__schema_name__: query_type for query_type in queryset_map}

        # Fetch all instances for each model, if some primary keys for it were returned in the union queryset.
        for typename, pk_to_index in pks_by_typename.items():
            query_type = query_type_by_name[typename]
            queryset = queryset_map[query_type]
            queryset = queryset.filter(pk__in=list(pk_to_index))
            instances = await evaluate_with_prefetch_hack_async(queryset)
            await self.check_permissions_async(root, info, query_type, instances)

            # Set the instances in the same order as the union query returned them.
            for instance in instances:
                pk = self.coalesce_pk(instance.pk)
                all_instances[pk_to_index[pk]] = instance

        return [instance for _, instance in sorted(all_instances.items())]

    def check_permissions(
        self,
        info: GQLInfo,
        root: Any,
        query_type: type[QueryType[TModel]],
        instances: list[TModel],
    ) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                self.entrypoint.permissions_func(root, info, instance)
            else:
                query_type.__permissions__(instance, info)

    async def check_permissions_async(
        self,
        info: GQLInfo,
        root: Any,
        query_type: type[QueryType[TModel]],
        instances: list[TModel],
    ) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                    await self.entrypoint.permissions_func(root, info, instance)
                else:
                    self.entrypoint.permissions_func(root, info, instance)

            elif inspect.iscoroutinefunction(query_type.__permissions__):
                await query_type.__permissions__(instance, info)

            else:
                query_type.__permissions__(instance, info)

    def optimize(self, info: GQLInfo, **kwargs: Any) -> QuerySetMap:
        queryset_map: QuerySetMap = {}

        for query_type in self.interface.__concrete_implementations__():
            model = query_type.__model__
            optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=model, info=info)
            optimizations = optimizer.compile()

            # If nothing is selected from this interface implementation, don't fetch anything from it
            if not optimizations:
                continue

            args: dict[str, Any] = {}
            filter_key = f"{undine_settings.QUERY_TYPE_FILTER_INPUT_KEY}{model.__name__}"
            order_by_key = f"{undine_settings.QUERY_TYPE_ORDER_INPUT_KEY}{model.__name__}"

            if filter_key in kwargs:
                args[undine_settings.QUERY_TYPE_FILTER_INPUT_KEY] = kwargs[filter_key]
            if order_by_key in kwargs:
                args[undine_settings.QUERY_TYPE_ORDER_INPUT_KEY] = kwargs[order_by_key]

            optimizer.handle_undine_query_type(query_type, args)

            optimizations.annotations["__typename"] = Value(query_type.__schema_name__)

            queryset = query_type.__get_queryset__(info)
            queryset_map[query_type] = optimizations.apply(queryset, info)

        return queryset_map

    def coalesce_pk(self, item: Any) -> Any:
        # Converting UUIDs to hex avoids an issue with union querysets with mixed int and UUID pks.
        return item.hex if isinstance(item, uuid.UUID) else item


@dataclasses.dataclass(frozen=True, slots=True)
class InterfaceTypeConnectionResolver(Generic[TModel]):
    """Resolves a connection of interface type items."""

    connection: Connection
    entrypoint: Entrypoint

    @property
    def interface_type(self) -> type[InterfaceType]:
        return self.connection.interface_type  # type: ignore[return-value]

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[ConnectionDict[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> ConnectionDict[TModel]:
        result = self.optimize(info, **kwargs)
        if not result.queryset_map:
            return self.empty_connection()

        all_instances = self.fetch_instances(root, info, result)
        return self.to_connection(all_instances, pagination=result.pagination)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> ConnectionDict[TModel]:
        result = self.optimize(info, **kwargs)
        if not result.queryset_map:
            return self.empty_connection()

        all_instances = await self.fetch_instances_async(root, info, result)
        return self.to_connection(all_instances, pagination=result.pagination)

    def fetch_instances(self, root: Any, info: GQLInfo, result: QuerySetMapWithPagination) -> list[TModel]:
        """
        Fetch paginated instances from the querysets.

        Use a union query to filter, order, and paginate the results together before fetching them.
        Then fetch the instances per-model and sort them in the same order as in the union query.
        """
        queryset_map = result.queryset_map
        pagination = result.pagination

        union_qs = create_union_queryset(queryset_map.values())
        union_qs = union_qs.order_by("pk")  # Default ordering

        # TODO: Filtering and ordering

        union_qs = union_qs.values("__typename", "pk")
        union_qs = pagination.paginate_queryset(union_qs, info)

        # Store the order in which the union query returned the items
        # from each model, so that we can sort when the actual instances are fetched.
        pks_by_typename: dict[str, dict[Hashable, int]] = defaultdict(dict)
        for index, item in enumerate(union_qs):
            pk = self.coalesce_pk(item["pk"])
            pks_by_typename[item["__typename"]][pk] = index

        all_instances: dict[int, TModel] = {}
        query_type_by_name = {query_type.__schema_name__: query_type for query_type in queryset_map}

        # Fetch all instances for each model, if some primary keys for it were returned in the union queryset.
        for typename, pk_to_index in pks_by_typename.items():
            query_type = query_type_by_name[typename]
            queryset = queryset_map[query_type]
            queryset = queryset.filter(pk__in=list(pk_to_index))
            instances = evaluate_with_prefetch_hack_sync(queryset)
            self.check_permissions(root, info, query_type, instances)

            # Set the instances in the same order as the union query returned them.
            for instance in instances:
                pk = self.coalesce_pk(instance.pk)
                all_instances[pk_to_index[pk]] = instance

        return [instance for _, instance in sorted(all_instances.items())]

    async def fetch_instances_async(self, root: Any, info: GQLInfo, result: QuerySetMapWithPagination) -> list[TModel]:
        """Async version of `fetch_instances`."""
        queryset_map = result.queryset_map
        pagination = result.pagination

        union_qs = create_union_queryset(queryset_map.values())
        union_qs = union_qs.order_by("pk")

        # TODO: Filtering and ordering

        union_qs = union_qs.values("__typename", "pk")
        # May call 'queryset.count()'.
        union_qs = await sync_to_async(pagination.paginate_queryset)(union_qs, info)

        # Store the order in which the union query returned the items
        # from each model, so that we can sort when the actual instances are fetched.
        pks_by_typename: dict[str, dict[Hashable, int]] = defaultdict(dict)
        counter = count()
        async for item in union_qs:
            pk = self.coalesce_pk(item["pk"])
            pks_by_typename[item["__typename"]][pk] = next(counter)

        all_instances: dict[int, TModel] = {}
        query_type_by_name = {query_type.__schema_name__: query_type for query_type in queryset_map}

        # Fetch all instances for each model, if some primary keys for it were returned in the union queryset.
        for typename, pk_to_index in pks_by_typename.items():
            query_type = query_type_by_name[typename]
            queryset = queryset_map[query_type]
            queryset = queryset.filter(pk__in=list(pk_to_index))
            instances = await evaluate_with_prefetch_hack_async(queryset)
            await self.check_permissions_async(root, info, query_type, instances)

            # Set the instances in the same order as the union query returned them.
            for instance in instances:
                pk = self.coalesce_pk(instance.pk)
                all_instances[pk_to_index[pk]] = instance

        return [instance for _, instance in sorted(all_instances.items())]

    def check_permissions(
        self,
        root: Any,
        info: GQLInfo,
        query_type: type[QueryType[TModel]],
        instances: list[TModel],
    ) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                self.entrypoint.permissions_func(root, info, instance)
            else:
                query_type.__permissions__(instance, info)

    async def check_permissions_async(
        self,
        root: Any,
        info: GQLInfo,
        query_type: type[QueryType[TModel]],
        instances: list[TModel],
    ) -> None:
        for instance in instances:
            if self.entrypoint.permissions_func is not None:
                if inspect.iscoroutinefunction(self.entrypoint.permissions_func):
                    await self.entrypoint.permissions_func(root, info, instance)
                else:
                    self.entrypoint.permissions_func(root, info, instance)

            elif inspect.iscoroutinefunction(query_type.__permissions__):
                await query_type.__permissions__(instance, info)

            else:
                query_type.__permissions__(instance, info)

    def optimize(self, info: GQLInfo, **kwargs: Any) -> QuerySetMapWithPagination:
        queryset_map: QuerySetMap = {}
        pagination: PaginationHandler | None = None

        for query_type in self.interface_type.__concrete_implementations__():
            model = query_type.__model__
            optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=model, info=info)
            optimizations = optimizer.compile()

            # If nothing is selected from this union type member, don't fetch anything from it
            if not optimizations:
                continue

            args: dict[str, Any] = {}
            filter_key = f"{undine_settings.QUERY_TYPE_FILTER_INPUT_KEY}{model.__name__}"
            order_by_key = f"{undine_settings.QUERY_TYPE_ORDER_INPUT_KEY}{model.__name__}"

            if filter_key in kwargs:
                args[undine_settings.QUERY_TYPE_FILTER_INPUT_KEY] = kwargs[filter_key]
            if order_by_key in kwargs:
                args[undine_settings.QUERY_TYPE_ORDER_INPUT_KEY] = kwargs[order_by_key]

            optimizer.handle_undine_query_type(query_type, args)

            optimizations.annotations["__typename"] = Value(query_type.__schema_name__)

            # Save pagination (should be the same for all query types)
            pagination = optimizations.pagination
            optimizations.pagination = None

            queryset = query_type.__get_queryset__(info)
            queryset_map[query_type] = optimizations.apply(queryset, info)

        return QuerySetMapWithPagination(queryset_map=queryset_map, pagination=pagination)

    def to_connection(self, instances: list[TModel], pagination: PaginationHandler) -> ConnectionDict[TModel]:
        typename = self.interface_type.__schema_name__
        edges = [
            NodeDict(
                cursor=offset_to_cursor(typename, pagination.start + index),
                node=instance,
            )
            for index, instance in enumerate(instances)
        ]
        return ConnectionDict(
            totalCount=pagination.total_count or 0,
            pageInfo=PageInfoDict(
                hasNextPage=(
                    False
                    if pagination.stop is None
                    else True
                    if pagination.total_count is None
                    else pagination.stop < pagination.total_count
                ),
                hasPreviousPage=pagination.start > 0,
                startCursor=None if not edges else edges[0]["cursor"],
                endCursor=None if not edges else edges[-1]["cursor"],
            ),
            edges=edges,
        )

    def empty_connection(self) -> ConnectionDict[TModel]:
        page_info = PageInfoDict(hasNextPage=False, hasPreviousPage=False, startCursor=None, endCursor=None)
        return ConnectionDict(totalCount=0, pageInfo=page_info, edges=[])

    def coalesce_pk(self, item: Any) -> Any:
        # Converting UUIDs to hex avoids an issue with union querysets with mixed int and UUID pks.
        return item.hex if isinstance(item, uuid.UUID) else item


# Miscellaneous


@dataclasses.dataclass(frozen=True, slots=True)
class TypedDictFieldResolver:
    """Resolves a typed dict field."""

    key: str
    """
    The actual key in the typed dict,
    which may have been converted to camel case for the GraphQL ObjectType.
    """

    def __call__(self, root: dict[str, Any], info: GQLInfo, **kwargs: Any) -> Any:
        return root.get(self.key)


@dataclasses.dataclass(frozen=True, slots=True)
class NamedTupleFieldResolver:
    """Resolves a named tuple field."""

    attr: str
    """
    The actual attribute in the named tuple,
    which may have been converted to camel case for the GraphQL ObjectType.
    """

    def __call__(self, root: object, info: GQLInfo, **kwargs: Any) -> Any:
        return getattr(root, self.attr, None)



from __future__ import annotations

import base64

from ._definition import ScalarType

__all__ = [
    "GraphQLBase16",
    "base16_scalar",
]


base16_scalar: ScalarType[str, str] = ScalarType(
    name="Base16",
    description="Represents a base16-encoded string.",
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc4648#section-8",
)

GraphQLBase16 = base16_scalar.as_graphql_scalar()


@base16_scalar.serialize.register
@base16_scalar.parse.register
def _(value: str) -> str:
    # Validate string is base16 encoded
    base64.b16decode(value.encode(encoding="utf-8")).decode(encoding="utf-8")
    return value


@base16_scalar.serialize.register
def _(value: bytes) -> str:
    # Validate string is base16 encoded
    base64.b16decode(value).decode(encoding="utf-8")
    return value.decode(encoding="utf-8")



from __future__ import annotations

import dataclasses
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, overload

from django.contrib.contenttypes.prefetch import GenericPrefetch
from django.db.models import ForeignKey, ManyToOneRel, OneToOneRel, Prefetch, Q
from django.db.models.constants import LOOKUP_SEP
from graphql import InlineFragmentNode, get_argument_values

from undine.converters import extend_expression
from undine.exceptions import GraphQLTooManyFiltersError, GraphQLTooManyOrdersError
from undine.settings import undine_settings
from undine.utils.graphql.undine_extensions import (
    get_undine_connection,
    get_undine_field,
    get_undine_mutation_type,
    get_undine_offset_pagination,
    get_undine_query_type,
)
from undine.utils.graphql.utils import get_underlying_type, is_typename_metafield, should_skip_node
from undine.utils.model_utils import get_default_manager, get_field_name, get_related_name
from undine.utils.reflection import is_same_func

from .ast_walker import GraphQLASTWalker
from .prefetch_hack import evaluate_with_prefetch_hack_async, evaluate_with_prefetch_hack_sync

if TYPE_CHECKING:
    from collections.abc import Generator

    from django.contrib.contenttypes.fields import GenericForeignKey
    from django.db.models import Field, Model, OrderBy, QuerySet
    from graphql import FieldNode, GraphQLInputObjectType, GraphQLInterfaceType, GraphQLObjectType, GraphQLScalarType

    from undine import Calculation, MutationType, QueryType
    from undine.pagination import PaginationHandler
    from undine.typing import (
        DjangoExpression,
        FilterCallback,
        GQLInfo,
        QuerySetCallback,
        RelatedField,
        Selections,
        TModel,
        ToManyField,
        ToOneField,
    )

__all__ = [
    "OptimizationData",
    "OptimizationResults",
    "QueryOptimizer",
    "optimize_async",
    "optimize_sync",
]


@overload
def optimize_sync(
    queryset: QuerySet[TModel],
    info: GQLInfo,
    *,
    offset: int = 0,
    limit: int | None = None,
) -> list[TModel]: ...


@overload
def optimize_sync(queryset: QuerySet[TModel], info: GQLInfo, **kwargs: Any) -> TModel | None: ...


def optimize_sync(
    queryset: QuerySet[TModel],
    info: GQLInfo,
    *,
    offset: int = 0,
    limit: int | None = None,
    **kwargs: Any,
) -> list[TModel] | TModel | None:
    """
    Optimize a queryset and return the results synchronously.

    :param queryset: The queryset to optimize.
    :param info: The GraphQL resolve info for the request.
    :param offset: The number of items to skip from the start. By default, no items are skipped.
    :param limit: The maximum number of items to return. By default, all items are returned.
    :param kwargs: Filtering that will result in a single item being returned.
    """
    optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=queryset.model, info=info)
    optimizations = optimizer.compile()
    optimized_queryset = optimizations.apply(queryset, info)

    if kwargs:
        optimized_queryset = optimized_queryset.filter(**kwargs)

    if limit is not None or offset > 0:
        optimized_queryset = optimized_queryset[offset : offset + limit]

    instances = evaluate_with_prefetch_hack_sync(optimized_queryset)

    if kwargs:
        return next(iter(instances), None)

    return instances


@overload
async def optimize_async(
    queryset: QuerySet[TModel],
    info: GQLInfo,
    *,
    limit: int | None = None,
    offset: int = 0,
) -> list[TModel]: ...


@overload
async def optimize_async(queryset: QuerySet[TModel], info: GQLInfo, **kwargs: Any) -> TModel | None: ...


async def optimize_async(
    queryset: QuerySet[TModel],
    info: GQLInfo,
    *,
    offset: int = 0,
    limit: int | None = None,
    **kwargs: Any,
) -> list[TModel] | TModel | None:
    """
    Optimize a queryset and return the results asynchronously.

    :param queryset: The queryset to optimize.
    :param info: The GraphQL resolve info for the request.
    :param offset: The number of items to skip from the start. By default, no items are skipped.
    :param limit: The maximum number of items to return. By default, all items are returned.
    :param kwargs: Filtering that will result in a single item being returned.
    """
    optimizer: QueryOptimizer = undine_settings.OPTIMIZER_CLASS(model=queryset.model, info=info)
    optimizations = optimizer.compile()
    optimized_queryset = optimizations.apply(queryset, info)

    if kwargs:
        optimized_queryset = optimized_queryset.filter(**kwargs)

    if limit is not None or offset > 0:
        optimized_queryset = optimized_queryset[offset : offset + limit]

    instances = await evaluate_with_prefetch_hack_async(optimized_queryset)

    if kwargs:
        return next(iter(instances), None)

    return instances


class QueryOptimizer(GraphQLASTWalker):
    """A class for processing the given GraphQL resolve info into required optimizations."""

    def __init__(self, *, model: type[Model], info: GQLInfo) -> None:
        """
        Optimize querysets based on the given GraphQL resolve info.

        :param model: The Django `Model` to start the optimization process from. Can be `None`
                      if the optimization is for a union of types.
        :param info: The GraphQL resolve info for the request. These are the "instructions" the optimizer follows
                     to compile the needed optimizations.
        """
        self.optimization_data = OptimizationData(model=model, info=info)
        super().__init__(info=info, model=model)

    def compile(self) -> OptimizationResults:
        self.run()
        return self.optimization_data.process()

    def parse_filter_info(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> None:
        """Parse filtering and ordering information from the given field."""
        graphql_field = parent_type.fields[field_node.name.value]
        object_type: GraphQLObjectType = get_underlying_type(graphql_field.type)  # type: ignore[assignment]

        arg_values = get_argument_values(graphql_field, field_node, self.info.variable_values)

        undine_connection = get_undine_connection(object_type)
        if undine_connection is not None:
            edge_type: GraphQLObjectType = get_underlying_type(object_type.fields["edges"].type)  # type: ignore[assignment]
            object_type = get_underlying_type(edge_type.fields["node"].type)  # type: ignore[assignment]

            self.optimization_data.pagination = undine_connection.pagination_handler(
                typename=object_type.name,
                first=arg_values.get("first"),
                last=arg_values.get("last"),
                after=arg_values.get("after"),
                before=arg_values.get("before"),
                page_size=undine_connection.page_size,
            )

        undine_offset_pagination = get_undine_offset_pagination(graphql_field)
        if undine_offset_pagination is not None:
            self.optimization_data.pagination = undine_offset_pagination.pagination_handler(
                typename=object_type.name,
                offset=arg_values.get("offset"),
                limit=arg_values.get("limit"),
                page_size=undine_offset_pagination.page_size,
            )

        # Check MutationType first so that it can override QueryType optimizations
        if parent_type == self.info.schema.mutation_type:
            arg = graphql_field.args.get(undine_settings.MUTATION_INPUT_DATA_KEY)
            if arg is not None:
                arg_type: GraphQLInputObjectType = get_underlying_type(arg.type)  # type: ignore[assignment]
                mutation_type = get_undine_mutation_type(arg_type)
                if mutation_type is not None:
                    self.handle_undine_mutation_type(mutation_type, arg_values)

        query_type = get_undine_query_type(object_type)
        if query_type is not None:
            self.handle_undine_query_type(query_type, arg_values)

    def handle_undine_mutation_type(self, mutation_type: type[MutationType], arg_values: dict[str, Any]) -> None:
        self.optimization_data.fill_from_mutation_type(mutation_type=mutation_type)

    def handle_undine_query_type(self, query_type: type[QueryType], arg_values: dict[str, Any]) -> None:
        self.optimization_data.fill_from_query_type(query_type=query_type)

        if query_type.__filterset__:
            filter_data = arg_values.get(undine_settings.QUERY_TYPE_FILTER_INPUT_KEY, {})
            filter_results = query_type.__filterset__.__build__(filter_data, self.info)

            if filter_results.filter_count > undine_settings.MAX_FILTERS_PER_TYPE:
                raise GraphQLTooManyFiltersError(
                    name=query_type.__filterset__.__schema_name__,
                    filter_count=filter_results.filter_count,
                    max_count=undine_settings.MAX_FILTERS_PER_TYPE,
                )

            self.optimization_data.filters.extend(filter_results.filters)
            self.optimization_data.aliases |= filter_results.aliases
            self.optimization_data.distinct |= filter_results.distinct
            self.optimization_data.none |= filter_results.none

        if query_type.__orderset__:
            order_data = arg_values.get(undine_settings.QUERY_TYPE_ORDER_INPUT_KEY, [])
            order_results = query_type.__orderset__.__build__(order_data, self.info)

            if order_results.order_count > undine_settings.MAX_ORDERS_PER_TYPE:
                raise GraphQLTooManyOrdersError(
                    name=query_type.__orderset__.__schema_name__,
                    filter_count=order_results.order_count,
                    max_count=undine_settings.MAX_ORDERS_PER_TYPE,
                )

            self.optimization_data.order_by.extend(order_results.order_by)
            self.optimization_data.aliases |= order_results.aliases

        query_type.__optimizations__(self.optimization_data, self.info)

    def handle_undine_field(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> None:
        """Handle any special operations for undine Fields."""
        graphql_field = parent_type.fields[field_node.name.value]
        undine_field = get_undine_field(graphql_field)
        if undine_field is None:
            return

        if undine_field.optimizer_func is not None:
            undine_field.optimizer_func(undine_field, self.optimization_data, self.info)

    def handle_query_class(self, field_type: GraphQLObjectType, field_node: FieldNode) -> None:
        self.parse_filter_info(field_type, field_node)
        super().handle_query_class(field_type, field_node)

    def handle_total_count(self, scalar: GraphQLScalarType, field_node: FieldNode) -> None:
        if self.optimization_data.pagination is not None:
            self.optimization_data.pagination.requires_total_count = True

    def handle_page_info_field(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> None:
        # To know if there is a next page, we must get total count.
        if self.optimization_data.pagination is not None and field_node.name.value != "hasNextPage":
            self.optimization_data.pagination.requires_total_count = True

    def handle_normal_field(self, parent_type: GraphQLObjectType, field_node: FieldNode, field: Field) -> None:
        # Aliases not accounted for since there is no filtering
        field_name = field.get_attname()
        self.optimization_data.only_fields.add(field_name)
        self.handle_undine_field(parent_type, field_node)

    def handle_to_one_field(
        self,
        parent_type: GraphQLObjectType,
        field_node: FieldNode,
        related_field: ToOneField,
    ) -> None:
        name = get_related_name(related_field)

        # Aliases not accounted for since there is no filtering
        data = self.optimization_data.add_select_related(name)

        if isinstance(related_field, ForeignKey):
            self.optimization_data.only_fields.add(related_field.attname)

        elif isinstance(related_field, OneToOneRel):
            data.only_fields.add(related_field.field.attname)

        self.handle_undine_field(parent_type, field_node)

        data.info = self.info
        with self.use_data(data):
            self.parse_filter_info(parent_type, field_node)
            super().handle_to_one_field(parent_type, field_node, related_field)

    def handle_to_many_field(
        self,
        parent_type: GraphQLObjectType,
        field_node: FieldNode,
        related_field: ToManyField,
    ) -> None:
        from django.contrib.contenttypes.fields import GenericRelation  # noqa: PLC0415

        name = get_field_name(related_field)
        alias = field_node.alias.value if field_node.alias else get_related_name(related_field)
        data = self.optimization_data.add_prefetch_related(name, to_attr=alias)

        if isinstance(related_field, ManyToOneRel):
            data.only_fields.add(related_field.field.attname)

        elif isinstance(related_field, GenericRelation):
            data.only_fields.add(related_field.object_id_field_name)
            data.only_fields.add(related_field.content_type_field_name)

        self.handle_undine_field(parent_type, field_node)

        data.info = self.info
        with self.use_data(data):
            self.parse_filter_info(parent_type, field_node)
            super().handle_to_many_field(parent_type, field_node, related_field)

    def handle_generic_foreign_key(
        self,
        parent_type: GraphQLObjectType,
        field_node: FieldNode,
        related_field: GenericForeignKey,
    ) -> None:
        name = get_related_name(related_field)

        self.optimization_data.only_fields.add(related_field.ct_field)
        self.optimization_data.only_fields.add(related_field.fk_field)

        self.handle_undine_field(parent_type, field_node)

        # GenericForeignKey should contain a selection set
        if field_node.selection_set is None:  # pragma: no cover
            return

        for selection in field_node.selection_set.selections:
            if should_skip_node(selection, self.info.variable_values):
                continue

            # GenericForeignKey can only contain InlineFragments for its related models,
            # or the __typename metafield
            if not isinstance(selection, InlineFragmentNode):
                if is_typename_metafield(field_node):
                    self.handle_typename_metafield(parent_type, field_node)

                continue

            fragment_name = selection.type_condition.name.value
            fragment_type: GraphQLObjectType = self.info.schema.get_type(fragment_name)  # type: ignore[assignment]
            fragment_model = self.get_model(fragment_type)

            # Only optimize union types that are query typ
            if fragment_model is None:  # pragma: no cover
                continue

            data = self.optimization_data.add_generic_prefetch_related(name, fragment_model)
            data.info = self.info

            with self.use_data(data):
                query_type = get_undine_query_type(fragment_type)
                if query_type is not None:
                    self.handle_undine_query_type(query_type, {})

                if selection.selection_set is None:
                    continue

                with self.use_model(fragment_model):
                    self.handle_selections(fragment_type, selection.selection_set.selections)

    def handle_node_interface(self, parent_type: GraphQLInterfaceType, selections: Selections) -> None:
        # Node ID might not be selected, but we'll still fetch it.
        self.optimization_data.only_fields.add(self.model._meta.pk.name)  # type: ignore[union-attr]

    def handle_custom_field(self, parent_type: GraphQLObjectType, field_node: FieldNode) -> None:
        self.handle_undine_field(parent_type, field_node)

    @contextmanager
    def use_data(self, nested_data: OptimizationData) -> Generator[None, Any, None]:
        original = self.optimization_data
        try:
            self.optimization_data = nested_data
            yield
        finally:
            self.optimization_data = original


@dataclasses.dataclass(slots=True)
class OptimizationData:
    """
    Holds QueryOptimizer optimization data. Can be processed to OptimizerResults
    when the optimization compilation is complete, which can then be used to optimize a queryset.
    """

    model: type[Model]
    info: GQLInfo

    parent: OptimizationData | None = None  # None if these are top-level optimizations.
    related_field: RelatedField | None = None  # Will be 'None' if there is no parent.

    only_fields: set[str] = dataclasses.field(default_factory=set)
    aliases: dict[str, DjangoExpression] = dataclasses.field(default_factory=dict)
    annotations: dict[str, DjangoExpression] = dataclasses.field(default_factory=dict)
    select_related: dict[str, OptimizationData] = dataclasses.field(default_factory=dict)
    prefetch_related: dict[str, OptimizationData] = dataclasses.field(default_factory=dict)
    generic_prefetches: dict[str, list[OptimizationData]] = dataclasses.field(default_factory=dict)

    filters: list[Q] = dataclasses.field(default_factory=list)
    order_by: list[OrderBy] = dataclasses.field(default_factory=list)
    distinct: bool = False
    none: bool = False
    pagination: PaginationHandler | None = None

    queryset_callback: QuerySetCallback = dataclasses.field(init=False)
    pre_filter_callback: FilterCallback | None = None
    post_filter_callback: FilterCallback | None = None
    field_calculations: list[Calculation] = dataclasses.field(default_factory=list)

    def __post_init__(self) -> None:
        def default_queryset_callback(info: GQLInfo) -> QuerySet:
            return get_default_manager(self.model).get_queryset()  # type: ignore[arg-type,union-attr]

        self.queryset_callback = default_queryset_callback

    def add_select_related(
        self,
        field_name: str,
        *,
        query_type: type[QueryType] | None = None,
    ) -> OptimizationData:
        """Add a 'select_related' optimization for the given field."""
        maybe_optimizer = self.select_related.get(field_name)
        if maybe_optimizer is not None:
            return maybe_optimizer

        field: ToOneField = self.model._meta.get_field(field_name)  # type: ignore[assignment,union-attr]
        model: type[Model] = field.related_model  # type: ignore[assignment]

        data = OptimizationData(model=model, info=self.info, related_field=field, parent=self)
        if query_type is not None:
            data.fill_from_query_type(query_type=query_type)

        self.select_related[field_name] = data

        # Fetch pk so that model is not "...both deferred and traversed using 'select_related' at the same time."
        data.only_fields.add(model._meta.pk.name)

        return data

    def add_prefetch_related(
        self,
        field_name: str,
        *,
        query_type: type[QueryType] | None = None,
        to_attr: str | None = None,
    ) -> OptimizationData:
        """Add a 'prefetch_related' optimization for the given field."""
        name = to_attr or field_name
        maybe_optimizer = self.prefetch_related.get(name)
        if maybe_optimizer is not None:
            return maybe_optimizer

        field: ToManyField = self.model._meta.get_field(field_name)  # type: ignore[assignment,union-attr]
        model: type[Model] = field.related_model  # type: ignore[assignment]

        data = OptimizationData(model=model, info=self.info, related_field=field, parent=self)
        if query_type is not None:
            data.fill_from_query_type(query_type=query_type)

        self.prefetch_related[name] = data
        return data

    def add_generic_prefetch_related(
        self,
        field_name: str,
        related_model: type[Model],
        *,
        query_type: type[QueryType] | None = None,
        to_attr: str | None = None,
    ) -> OptimizationData:
        """Add a 'prefetch_related' optimization for a generic prefetch for the given model."""
        field: ToManyField = self.model._meta.get_field(field_name)  # type: ignore[assignment,union-attr]

        name = to_attr or field_name
        prefetch_data: list[OptimizationData] = self.generic_prefetches.setdefault(name, [])

        data: OptimizationData | None = None

        if prefetch_data:
            data = next((o for o in prefetch_data if o.model == related_model), None)

        if data is None:
            data = OptimizationData(model=related_model, info=self.info, related_field=field, parent=self)
            prefetch_data.append(data)

        if query_type is not None:
            data.fill_from_query_type(query_type=query_type)

        return data

    def should_promote_to_prefetch(self) -> bool:
        """
        If this is a `to-one` related field on some other field, should its results be fetched with
        `prefetch_related` instead of `select_related`?

        E.g. If the model instances need to be annotated, we need to prefetch to retain those annotations.
        Or if we have filtering that always needs to be done on the related objects.
        """
        return (
            bool(self.annotations)
            or bool(self.aliases)
            or bool(self.field_calculations)
            or self.pre_filter_callback is not None
            or self.post_filter_callback is not None
        )

    def fill_from_query_type(self, query_type: type[QueryType]) -> OptimizationData:
        """Fill the optimization data from the given QueryType."""
        from undine import FilterSet, QueryType  # noqa: PLC0415

        self.model = query_type.__model__
        self.queryset_callback = query_type.__get_queryset__

        # Only include pre-filter callback if it's different from the default.
        if (
            self.pre_filter_callback is None  # No custom pre-filter callback
            and not is_same_func(query_type.__filter_queryset__, QueryType.__filter_queryset__)
        ):
            self.pre_filter_callback = query_type.__filter_queryset__

        # Only include post-filter callback if it's different from the default.
        if (
            self.post_filter_callback is None  # No custom post-filter callback
            and query_type.__filterset__  # Has filterset
            and not is_same_func(query_type.__filterset__.__filter_queryset__, FilterSet.__filter_queryset__)
        ):
            self.post_filter_callback = query_type.__filterset__.__filter_queryset__

        return self

    def fill_from_mutation_type(self, mutation_type: type[MutationType]) -> OptimizationData:
        """Fill the optimization data from the given MutationType."""
        from undine import MutationType  # noqa: PLC0415

        # Only include pre-filter callback if it's different from the default.
        if (
            self.pre_filter_callback is None  # No custom pre-filter callback
            and not is_same_func(mutation_type.__filter_queryset__, MutationType.__filter_queryset__)
        ):
            self.pre_filter_callback = mutation_type.__filter_queryset__

        return self

    def process(self) -> OptimizationResults:
        """Process collected data to OptimizerResults that can be applied to a queryset."""
        results = OptimizationResults(
            related_field=self.related_field,
            only_fields=self.only_fields,
            aliases=self.aliases,
            annotations=self.annotations,
            filters=self.filters,
            order_by=self.order_by,
            distinct=self.distinct,
            none=self.none,
            pagination=self.pagination,
            pre_filter_callback=self.pre_filter_callback,
            post_filter_callback=self.post_filter_callback,
            field_calculations=self.field_calculations,
        )

        for select_related_data in self.select_related.values():
            # Check if we need to prefetch instead of joining.
            if select_related_data.should_promote_to_prefetch():
                prefetch = self._process_prefetch(select_related_data)
                results.prefetch_related.add(prefetch)
                continue

            # Otherwise extend lookups to this model.
            nested_results = select_related_data.process()
            results.extend(nested_results)

        for name, prefetch_related_data in self.prefetch_related.items():
            # Only need `to_attr` if it's different from the field name.
            to_attr: str | None = name
            if to_attr == get_related_name(prefetch_related_data.related_field):  # type: ignore[union-attr]
                to_attr = None

            prefetch = self._process_prefetch(prefetch_related_data, to_attr=to_attr)
            results.prefetch_related.add(prefetch)

        for name, generic_prefetches in self.generic_prefetches.items():
            if not generic_prefetches:
                continue

            # Only need `to_attr` if it's different from the field name.
            to_attr = name
            if to_attr == get_related_name(generic_prefetches[0].related_field):  # type: ignore[union-attr]
                to_attr = None

            generic_prefetch = self._process_generic_prefetch(generic_prefetches, to_attr=to_attr)
            results.prefetch_related.add(generic_prefetch)

        return results

    def _process_prefetch(self, data: OptimizationData, *, to_attr: str | None = None) -> Prefetch:
        """Process prefetch related optimization data to a Prefetch object."""
        optimizations = data.process()
        queryset = data.queryset_callback(data.info)  # type: ignore[arg-type]
        optimized_queryset = optimizations.apply(queryset, data.info)  # type: ignore[arg-type]

        field_name = get_related_name(data.related_field)  # type: ignore[union-attr]
        return Prefetch(field_name, optimized_queryset, to_attr=to_attr)

    def _process_generic_prefetch(self, data: list[OptimizationData], *, to_attr: str | None = None) -> GenericPrefetch:
        """Process generic foreign key optimization data to a GenericPrefetch object."""
        optimized_querysets: list[QuerySet] = []

        for model_data in data:
            optimizations = model_data.process()
            queryset = model_data.queryset_callback(model_data.info)  # type: ignore[arg-type]
            optimized_queryset = optimizations.apply(queryset, model_data.info)  # type: ignore[arg-type]
            optimized_querysets.append(optimized_queryset)

        field_name = get_related_name(data[0].related_field)  # type: ignore[union-attr]
        return GenericPrefetch(field_name, optimized_querysets, to_attr=to_attr)


@dataclasses.dataclass(slots=True)
class OptimizationResults:
    """Optimizations that can be applied to a queryset."""

    # The model field for a relation, if these are the results for a prefetch.
    related_field: RelatedField | None = None

    # Field optimizations
    only_fields: set[str] = dataclasses.field(default_factory=set)
    aliases: dict[str, DjangoExpression] = dataclasses.field(default_factory=dict)
    annotations: dict[str, DjangoExpression] = dataclasses.field(default_factory=dict)
    select_related: set[str] = dataclasses.field(default_factory=set)
    prefetch_related: set[Prefetch | str] = dataclasses.field(default_factory=set)

    # Filtering
    filters: list[Q] = dataclasses.field(default_factory=list)
    order_by: list[OrderBy] = dataclasses.field(default_factory=list)
    distinct: bool = False
    none: bool = False
    pagination: PaginationHandler | None = None

    pre_filter_callback: FilterCallback | None = None
    post_filter_callback: FilterCallback | None = None
    field_calculations: list[Calculation] = dataclasses.field(default_factory=list)

    def apply(self, queryset: QuerySet[TModel], info: GQLInfo) -> QuerySet[TModel]:  # noqa: C901, PLR0912
        """Apply the optimization results to the given queryset."""
        if self.none:
            return queryset.none()

        if self.select_related:
            queryset = queryset.select_related(*self.select_related)
        if self.prefetch_related:
            queryset = queryset.prefetch_related(*self.prefetch_related)
        if not undine_settings.DISABLE_ONLY_FIELDS_OPTIMIZATION and self.only_fields:
            queryset = queryset.only(*self.only_fields)
        if self.aliases:
            queryset = queryset.alias(**self.aliases)
        if self.annotations:
            queryset = queryset.annotate(**self.annotations)

        if self.pre_filter_callback is not None:
            queryset = self.pre_filter_callback(queryset, info)

        if self.order_by:
            queryset = queryset.order_by(*self.order_by)
        if self.distinct:
            queryset = queryset.distinct()

        for calculation in self.field_calculations:
            queryset = queryset.annotate(**{calculation.__field_name__: calculation(info)})

        # Note that we want to add the filters as as single Q object to prevent some issues
        # when filters are "spanning multi-valued relationships". See Django documentation here:
        # https://docs.djangoproject.com/en/stable/topics/db/queries/#spanning-multi-valued-relationships
        #
        # Specifically, if two filters for the same many-valued relationships are used in separate
        # `queryset.filter(...)` calls, the result of the filtering will include results where EITHER
        # of the conditions are met, not BOTH. Furthermore, separate `queryset.filter(...)` calls
        # will result in multiple joins to the many-valued relationship, which can be very expensive.
        # The EITHER behavior is still possible by using an OR block of a FilterSet.
        if self.filters:
            queryset = queryset.filter(Q(*self.filters))

        if self.post_filter_callback is not None:
            queryset = self.post_filter_callback(queryset, info)

        if self.pagination is not None:
            if self.related_field is None:
                queryset = self.pagination.paginate_queryset(queryset, info)
            else:
                queryset = self.pagination.paginate_prefetch_queryset(queryset, self.related_field, info)  # type: ignore[arg-type]

        return queryset

    def extend(self, other: OptimizationResults) -> OptimizationResults:
        """
        Extend the given optimization results to this one
        by prefixing their lookups using the other's `field_name`.
        """
        self.select_related.add(other.related_field.name)  # type: ignore[union-attr]
        self.only_fields.update(f"{other.related_field.name}{LOOKUP_SEP}{only}" for only in other.only_fields)  # type: ignore[union-attr]
        self.select_related.update(f"{other.related_field.name}{LOOKUP_SEP}{select}" for select in other.select_related)  # type: ignore[union-attr]

        for prefetch in other.prefetch_related:
            if isinstance(prefetch, str):
                self.prefetch_related.add(f"{other.related_field.name}{LOOKUP_SEP}{prefetch}")  # type: ignore[union-attr]
            if isinstance(prefetch, Prefetch):
                prefetch.add_prefix(other.related_field.name)  # type: ignore[union-attr]
                self.prefetch_related.add(prefetch)

        self.filters.extend(extend_expression(ftr, field_name=other.related_field.name) for ftr in other.filters)  # type: ignore[union-attr,misc]
        self.order_by.extend(extend_expression(order, field_name=other.related_field.name) for order in other.order_by)  # type: ignore[union-attr,misc]
        self.distinct |= other.distinct
        self.none |= other.none

        return self

    def __bool__(self) -> bool:
        return bool(self.only_fields or self.annotations or self.select_related or self.prefetch_related)



from __future__ import annotations

import itertools
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRel
from django.db.models import F, Model
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.query_utils import DeferredAttribute, Q
from graphql import GraphQLInputType

from undine import Order
from undine.converters import convert_to_graphql_type, convert_to_order_ref
from undine.exceptions import UnionModelFieldDirectUsageError, UnionModelFieldMismatchError
from undine.typing import CombinableExpression, ModelField
from undine.utils.model_utils import determine_output_field, get_model_field


@convert_to_order_ref.register
def _(ref: str, **kwargs: Any) -> Any:
    caller: Order = kwargs["caller"]

    models = caller.orderset.__models__
    if len(models) == 1:
        get_model_field(model=caller.orderset.__models__[0], lookup=ref)
        return F(ref)

    fields_by_model: dict[type[Model], GraphQLInputType] = {}
    for model in caller.orderset.__models__:
        field = get_model_field(model=model, lookup=caller.field_name)
        fields_by_model[model] = convert_to_graphql_type(field, model=model, is_input=True)

    for (model_1, field_1), (model_2, field_2) in itertools.combinations(fields_by_model.items(), 2):
        if field_1 != field_2:
            raise UnionModelFieldMismatchError(
                ref=ref,
                type_1=field_1,
                model_1=model_1,
                type_2=field_2,
                model_2=model_2,
                kind="Order",
            )

    return F(ref)


@convert_to_order_ref.register
def _(_: None, **kwargs: Any) -> Any:
    caller: Order = kwargs["caller"]
    return convert_to_order_ref(caller.field_name, **kwargs)


@convert_to_order_ref.register
def _(ref: F | Q, **kwargs: Any) -> Any:
    return ref


@convert_to_order_ref.register
def _(ref: ModelField, **kwargs: Any) -> Any:
    caller = kwargs["caller"]

    models = caller.orderset.__models__
    if len(models) != 1:
        raise UnionModelFieldDirectUsageError(kind="OrderSet")

    return F(ref.name)


@convert_to_order_ref.register
def _(ref: CombinableExpression, **kwargs: Any) -> Any:
    caller: Order = kwargs["caller"]
    models = caller.orderset.__models__

    if len(models) == 1:
        ref.output_field = determine_output_field(ref, model=caller.orderset.__models__[0])
        return ref

    fields_by_model: dict[type[Model], ModelField] = {}
    for model in models:
        fields_by_model[model] = determine_output_field(ref, model=model)

    for (model_1, field_1), (model_2, field_2) in itertools.combinations(fields_by_model.items(), 2):
        if field_1.__class__ is not field_2.__class__:
            raise UnionModelFieldMismatchError(
                ref=ref,
                type_1=field_1,
                model_1=model_1,
                type_2=field_2,
                model_2=model_2,
                kind="Order",
            )

    ref.output_field = next(iter(fields_by_model.values()))
    return ref


@convert_to_order_ref.register
def _(ref: DeferredAttribute | ForwardManyToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_order_ref(ref.field, **kwargs)


@convert_to_order_ref.register
def _(ref: ReverseManyToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_order_ref(ref.rel, **kwargs)


@convert_to_order_ref.register
def _(ref: ReverseOneToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_order_ref(ref.related, **kwargs)


@convert_to_order_ref.register
def _(ref: ManyToManyDescriptor, **kwargs: Any) -> Any:
    return convert_to_order_ref(ref.rel if ref.reverse else ref.field, **kwargs)


@convert_to_order_ref.register
def _(ref: GenericRel, **kwargs: Any) -> Any:
    return convert_to_order_ref(ref.field, **kwargs)


@convert_to_order_ref.register  # Required for Django<5.1
def _(ref: GenericForeignKey, **kwargs: Any) -> Any:
    caller = kwargs["caller"]

    models = caller.orderset.__models__
    if len(models) != 1:
        raise UnionModelFieldDirectUsageError(kind="OrderSet")

    return F(ref.name)



from __future__ import annotations

import dataclasses
from types import SimpleNamespace
from typing import TYPE_CHECKING, Any, Generic

from asgiref.sync import sync_to_async
from django.db.models import Model, Q
from graphql import Undefined

from undine.exceptions import GraphQLMissingLookupFieldError, GraphQLMutationInstanceLimitError
from undine.settings import undine_settings
from undine.typing import TModel
from undine.utils.graphql.utils import pre_evaluate_request_user
from undine.utils.model_utils import (
    convert_integrity_errors,
    get_default_manager,
    get_instance_or_raise,
    get_instances_or_raise,
    get_pks_from_list_of_dicts,
)
from undine.utils.pre_mutation import pre_mutation, pre_mutation_async, pre_mutation_many, pre_mutation_many_async
from undine.utils.reflection import as_coroutine_func_if_not

from .query import QueryTypeManyResolver, QueryTypeSingleResolver

if TYPE_CHECKING:
    from graphql.pyutils import AwaitableOrValue

    from undine import Entrypoint, GQLInfo, MutationType, QueryType

__all__ = [
    "BulkCreateResolver",
    "BulkDeleteResolver",
    "BulkUpdateResolver",
    "CreateResolver",
    "DeleteResolver",
    "UpdateResolver",
]


# Single


@dataclasses.dataclass(frozen=True, slots=True)
class CreateResolver(Generic[TModel]):
    """Resolves a mutation for creating a model instance using."""

    mutation_type: type[MutationType[TModel]]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[TModel | None]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    @property
    def query_type(self) -> type[QueryType[TModel]]:
        return self.mutation_type.__query_type__()

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        input_data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        instance = self.model()
        pre_mutation(
            instance=instance,
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        with convert_integrity_errors():
            instance = self.mutation_type.__mutate__(instance=instance, info=info, input_data=input_data)

        if isinstance(instance, Model):
            self.mutation_type.__after__(instance=instance, info=info, input_data=input_data)  # type: ignore[arg-type]

            resolver = QueryTypeSingleResolver(query_type=self.query_type, entrypoint=self.entrypoint)
            return resolver.run_sync(root, info, pk=instance.pk)

        return instance

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        input_data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        instance = self.model()
        await pre_mutation_async(
            instance=instance,
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        mutate_func = as_coroutine_func_if_not(self.mutation_type.__mutate__)

        with convert_integrity_errors():
            instance = await mutate_func(instance=instance, info=info, input_data=input_data)

        if isinstance(instance, Model):
            after_func = as_coroutine_func_if_not(self.mutation_type.__after__)
            await after_func(instance=instance, info=info, input_data=input_data)

            resolver = QueryTypeSingleResolver(query_type=self.query_type, entrypoint=self.entrypoint)
            return await resolver.run_async(root, info, pk=instance.pk)

        return instance


@dataclasses.dataclass(frozen=True, slots=True)
class UpdateResolver(Generic[TModel]):
    """Resolves a mutation for updating a model instance."""

    mutation_type: type[MutationType[TModel]]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[TModel | None]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    @property
    def query_type(self) -> type[QueryType[TModel]]:
        return self.mutation_type.__query_type__()

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        input_data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        if "pk" not in input_data:
            raise GraphQLMissingLookupFieldError(model=self.model, key="pk")

        instance = get_instance_or_raise(model=self.model, pk=input_data["pk"])

        pre_mutation(
            instance=instance,
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        with convert_integrity_errors():
            instance = self.mutation_type.__mutate__(instance=instance, info=info, input_data=input_data)

        if isinstance(instance, Model):
            self.mutation_type.__after__(instance=instance, info=info, input_data=input_data)  # type: ignore[arg-type]

            resolver = QueryTypeSingleResolver(query_type=self.query_type, entrypoint=self.entrypoint)
            return resolver.run_sync(root, info, pk=instance.pk)

        return instance

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> TModel | None:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        input_data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        if "pk" not in input_data:
            raise GraphQLMissingLookupFieldError(model=self.model, key="pk")

        instance = await sync_to_async(get_instance_or_raise)(model=self.model, pk=input_data["pk"])
        await pre_mutation_async(
            instance=instance,
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        mutate_func = as_coroutine_func_if_not(self.mutation_type.__mutate__)

        with convert_integrity_errors():
            instance = await mutate_func(instance=instance, info=info, input_data=input_data)

        if isinstance(instance, Model):
            after_func = as_coroutine_func_if_not(self.mutation_type.__after__)
            await after_func(instance=instance, info=info, input_data=input_data)

            resolver = QueryTypeSingleResolver(query_type=self.query_type, entrypoint=self.entrypoint)
            return await resolver.run_async(root, info, pk=instance.pk)

        return instance


@dataclasses.dataclass(frozen=True, slots=True)
class DeleteResolver(Generic[TModel]):
    """Resolves a mutation for deleting a model instance."""

    mutation_type: type[MutationType]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[SimpleNamespace]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> SimpleNamespace:
        input_data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        pk: Any = input_data.get("pk", Undefined)
        if pk is Undefined:
            raise GraphQLMissingLookupFieldError(model=self.model, key="pk")

        instance = get_instance_or_raise(model=self.model, pk=input_data["pk"])

        pre_mutation(
            instance=instance,
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        with convert_integrity_errors():
            instance.delete()

        self.mutation_type.__after__(instance=instance, info=info, input_data=input_data)

        return SimpleNamespace(pk=pk)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> SimpleNamespace:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        input_data: dict[str, Any] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        pk: Any = input_data.get("pk", Undefined)
        if pk is Undefined:
            raise GraphQLMissingLookupFieldError(model=self.model, key="pk")

        instance = await sync_to_async(get_instance_or_raise)(model=self.model, pk=input_data["pk"])

        await pre_mutation_async(
            instance=instance,
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        with convert_integrity_errors():
            await instance.adelete()

        after_func = as_coroutine_func_if_not(self.mutation_type.__after__)
        await after_func(instance=instance, info=info, input_data=input_data)

        return SimpleNamespace(pk=pk)


# Bulk


@dataclasses.dataclass(frozen=True, slots=True)
class BulkCreateResolver(Generic[TModel]):
    """Resolves a bulk create mutation for creating a list of model instances."""

    mutation_type: type[MutationType[TModel]]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    @property
    def query_type(self) -> type[QueryType[TModel]]:
        return self.mutation_type.__query_type__()

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        input_data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        count = len(input_data)
        if count > undine_settings.MUTATION_INSTANCE_LIMIT:
            raise GraphQLMutationInstanceLimitError(limit=undine_settings.MUTATION_INSTANCE_LIMIT, count=count)

        instances: list[TModel] = [self.model() for _ in input_data]

        pre_mutation_many(
            instances=instances,  # type: ignore[arg-type]
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        with convert_integrity_errors():
            instances = self.mutation_type.__bulk_mutate__(instances=instances, info=info, input_data=input_data)

        for instance, data in zip(instances, input_data, strict=True):
            self.mutation_type.__after__(instance=instance, info=info, input_data=data)

        resolver = QueryTypeManyResolver(
            query_type=self.query_type,
            entrypoint=self.entrypoint,
            additional_filter=Q(pk__in=[instance.pk for instance in instances]),
        )
        return resolver.run_sync(root, info)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        input_data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        count = len(input_data)
        if count > undine_settings.MUTATION_INSTANCE_LIMIT:
            raise GraphQLMutationInstanceLimitError(limit=undine_settings.MUTATION_INSTANCE_LIMIT, count=count)

        instances: list[TModel] = [self.model() for _ in input_data]

        await pre_mutation_many_async(
            instances=instances,  # type: ignore[arg-type]
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        mutate_func = as_coroutine_func_if_not(self.mutation_type.__bulk_mutate__)

        with convert_integrity_errors():
            instances = await mutate_func(instances=instances, info=info, input_data=input_data)

        after_func = as_coroutine_func_if_not(self.mutation_type.__after__)

        for instance, data in zip(instances, input_data, strict=True):
            await after_func(instance=instance, info=info, input_data=data)

        resolver = QueryTypeManyResolver(
            query_type=self.query_type,
            entrypoint=self.entrypoint,
            additional_filter=Q(pk__in=[instance.pk for instance in instances]),
        )
        return await resolver.run_async(root, info)


@dataclasses.dataclass(frozen=True, slots=True)
class BulkUpdateResolver(Generic[TModel]):
    """Resolves a bulk update mutation for updating a list of model instances."""

    mutation_type: type[MutationType[TModel]]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[TModel]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    @property
    def query_type(self) -> type[QueryType[TModel]]:
        return self.mutation_type.__query_type__()

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        input_data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        count = len(input_data)
        if count > undine_settings.MUTATION_INSTANCE_LIMIT:
            raise GraphQLMutationInstanceLimitError(limit=undine_settings.MUTATION_INSTANCE_LIMIT, count=count)

        pks = get_pks_from_list_of_dicts(input_data)
        instances = get_instances_or_raise(model=self.model, pks=pks)

        pre_mutation_many(
            instances=instances,  # type: ignore[arg-type]
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        with convert_integrity_errors():
            instances = self.mutation_type.__bulk_mutate__(instances=instances, info=info, input_data=input_data)

        for instance, data in zip(instances, input_data, strict=True):
            self.mutation_type.__after__(instance=instance, info=info, input_data=data)

        resolver = QueryTypeManyResolver(
            query_type=self.query_type,
            entrypoint=self.entrypoint,
            additional_filter=Q(pk__in=[instance.pk for instance in instances]),
        )
        return resolver.run_sync(root, info)

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[TModel]:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        input_data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        count = len(input_data)
        if count > undine_settings.MUTATION_INSTANCE_LIMIT:
            raise GraphQLMutationInstanceLimitError(limit=undine_settings.MUTATION_INSTANCE_LIMIT, count=count)

        pks = get_pks_from_list_of_dicts(input_data)
        instances = await sync_to_async(get_instances_or_raise)(model=self.model, pks=pks)

        await pre_mutation_many_async(
            instances=instances,  # type: ignore[arg-type]
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        mutate_func = as_coroutine_func_if_not(self.mutation_type.__bulk_mutate__)

        with convert_integrity_errors():
            instances = await mutate_func(instances=instances, info=info, input_data=input_data)

        after_func = as_coroutine_func_if_not(self.mutation_type.__after__)

        for instance, data in zip(instances, input_data, strict=True):
            await after_func(instance=instance, info=info, input_data=data)

        resolver = QueryTypeManyResolver(
            query_type=self.query_type,
            entrypoint=self.entrypoint,
            additional_filter=Q(pk__in=[instance.pk for instance in instances]),
        )
        return await resolver.run_async(root, info)


@dataclasses.dataclass(frozen=True, slots=True)
class BulkDeleteResolver(Generic[TModel]):
    """Resolves a bulk delete mutation for deleting a list of model instances."""

    mutation_type: type[MutationType]
    entrypoint: Entrypoint

    def __call__(self, root: Any, info: GQLInfo, **kwargs: Any) -> AwaitableOrValue[list[SimpleNamespace]]:
        if undine_settings.ASYNC:
            return self.run_async(root, info, **kwargs)
        return self.run_sync(root, info, **kwargs)

    @property
    def model(self) -> type[TModel]:
        return self.mutation_type.__model__  # type: ignore[return-value]

    def run_sync(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[SimpleNamespace]:
        input_data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        count = len(input_data)
        if count > undine_settings.MUTATION_INSTANCE_LIMIT:
            raise GraphQLMutationInstanceLimitError(limit=undine_settings.MUTATION_INSTANCE_LIMIT, count=count)

        pks = get_pks_from_list_of_dicts(input_data)
        instances = get_instances_or_raise(model=self.model, pks=pks)

        pre_mutation_many(
            instances=instances,  # type: ignore[arg-type]
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        with convert_integrity_errors():
            get_default_manager(self.model).filter(pk__in=pks).delete()

        for instance, data in zip(instances, input_data, strict=True):
            self.mutation_type.__after__(instance=instance, info=info, input_data=data)

        return [SimpleNamespace(pk=pk) for pk in pks]

    async def run_async(self, root: Any, info: GQLInfo, **kwargs: Any) -> list[SimpleNamespace]:
        # Fetch user eagerly so that its available in synchronous parts of the code.
        await pre_evaluate_request_user(info)

        input_data: list[dict[str, Any]] = kwargs[undine_settings.MUTATION_INPUT_DATA_KEY]

        pks = get_pks_from_list_of_dicts(input_data)
        instances = await sync_to_async(get_instances_or_raise)(model=self.model, pks=pks)

        await pre_mutation_many_async(
            instances=instances,  # type: ignore[arg-type]
            info=info,
            input_data=input_data,
            mutation_type=self.mutation_type,
        )

        with convert_integrity_errors():
            await get_default_manager(self.model).filter(pk__in=pks).adelete()

        after_func = as_coroutine_func_if_not(self.mutation_type.__after__)

        for instance, data in zip(instances, input_data, strict=True):
            await after_func(instance=instance, info=info, input_data=data)

        return [SimpleNamespace(pk=pk) for pk in pks]



from __future__ import annotations

import inspect
from types import FunctionType
from typing import Any

from graphql import GraphQLFieldResolver

from undine import Entrypoint
from undine.converters import convert_to_entrypoint_subscription
from undine.resolvers import FunctionSubscriptionResolver
from undine.resolvers.subscription import ModelDeleteSubscriptionResolver, ModelSaveSubscriptionResolver
from undine.subscriptions import ModelDeleteSubscription, ModelSaveSubscription


@convert_to_entrypoint_subscription.register
def _(_: Any, **kwargs: Any) -> GraphQLFieldResolver | None:
    # Don't create a subscription resolver for any Entrypoint reference that is not used for subscriptions
    return None


@convert_to_entrypoint_subscription.register
def _(ref: FunctionType, **kwargs: Any) -> GraphQLFieldResolver | None:
    if not inspect.isasyncgenfunction(ref) and not inspect.iscoroutinefunction(ref):
        return None

    # We don't know if the function submitted here is actually for a subscription,
    # or if it returns a something that can be used for subscriptions,
    # but there is no harm in creating the resolver anyway.
    caller: Entrypoint = kwargs["caller"]
    return FunctionSubscriptionResolver(func=ref, entrypoint=caller)


@convert_to_entrypoint_subscription.register
def _(ref: ModelSaveSubscription, **kwargs: Any) -> GraphQLFieldResolver | None:
    caller: Entrypoint = kwargs["caller"]
    return ModelSaveSubscriptionResolver(subscription=ref, entrypoint=caller)


@convert_to_entrypoint_subscription.register
def _(ref: ModelDeleteSubscription, **kwargs: Any) -> GraphQLFieldResolver | None:
    caller: Entrypoint = kwargs["caller"]
    return ModelDeleteSubscriptionResolver(subscription=ref, entrypoint=caller)
